#!/bin/sh

PGW="pgw"
LIB="libptr"
PTR="pointer"
TEST="test"

XROOT=""
CC=""
INSTDIR="/usr/local"
USE_LIB_XPM="no"
USE_LIB_GIF="no"
USE_LIB_UNGIF="no"


if [ -f "./config.log" ]; then
    eval `cat ./config.log`
fi

RANLIB="yes"
CTOF="need"
USE_SELECT="succeeded"
USE_OPENDIR="succeeded"
USE_SCANDIR="succeeded"


XPATHES="/usr/X11R6 /usr/local/X11R5 /usr/local/X11 /usr/X11 /usr/openwin"
BINPATH=`echo "$PATH" | tr ':' ' '`
LIBPATH=". $HOME/lib /usr/local/lib /usr/lib"

# Use Sysv echo if possible
if [ -x /usr/5bin/echo ]       # SunOS SysV echo
then
    ECHON=/usr/5bin/echo
elif [ -z "`(echo -e a) 2>/dev/null | grep e`" ]  # GNU bash, etc.
then
    ECHON="echo -e"
else                           # generic SysV
    ECHON=echo
fi



# utility: search path
searchpath()
{
   file=$1
   shift
   for i
   do
       if [ -f $i/$file ]; then
           echo $i
           return
       elif [ -d $i/$file ]; then
           echo $i
           return
       fi
   done
}

# utility: ask the question
ask_str()
{
    $ECHON "$*\c"
    read TMP
}



# C compiler
if [ "$CC" = "" ]; then
    TMP=`searchpath gcc $BINPATH`
    if [ "$TMP" = "" ]; then
        CC="cc -O"
    else
        CC="gcc -O"
    fi
fi
ask_str "C compiler [$CC]: "
if [ "$TMP" != "" ]; then
    CC="$TMP"
fi


# check arguments
for av in $*; do
    case "$av" in
      --enable-xpm)
        USE_LIB_XPM="yes";;
      --disable-xpm)
        USE_LIB_XPM="no";;
      --enable-gif)
        USE_LIB_GIF="yes";;
      --disable-gif)
        USE_LIB_GIF="no";;
      --enable-ungif)
        USE_LIB_UNGIF="yes";;
      --disable-ungif)
        USE_LIB_UNGIF="no";;
    esac
done


# install directory
ask_str "install directory [$INSTDIR]: "
if [ "$TMP" != "" ]; then
    INSTDIR="$TMP"
fi
INSTDIR=`echo "$INSTDIR" | sed "s!~!$HOME!"`


# BSD or SYSV
if [ "$OS" = "" ]; then
    OS="SYSV"
    TMP=`uname -a | grep '[bB][sS][dD]'`
    if [ "$TMP" != "" ]; then
        OS="BSD"
    fi
fi
ask_str "SYSV or BSD [$OS]: "
if [ "$TMP" != "" ]; then
    OS="$TMP"
fi

# search Xlib path
if [ "$XROOT" = "" ]; then
    XROOT=`searchpath xterm $BINPATH`
fi
if [ "$XROOT" = "" ]; then
    XROOT=`searchpath include/X11 $XPATHES`
else
    XROOT=`echo $XROOT | sed "s!/bin!!"`
fi
ask_str "X root directory [$XROOT]: "
if [ "$TMP" != "" ]; then
    XROOT="$TMP"
fi
if [ "$XROOT" != "" ]; then
    echo "X root directory: $XROOT"
fi
while [ "$XROOT" = "" ]; do
    ask_str "X root directory [$XROOT]: "
    if [ "$TMP" != "" ]; then
        XROOT="$TMP"
    fi
done



# generate "$TEST/Makefile" from "$TEST/Makefile.in"
TMP=sedscript_$$
cat >$TMP << _EOF_
s!XROOT = .*!XROOT = $XROOT!
s!CC = .*!CC = $CC!
s!OS = .*!OS = $OS!
_EOF_

if [ -f $TEST/Makefile ]; then
    mv -f $TEST/Makefile $TEST/Makefile.bac
fi
sed -f $TMP $TEST/Makefile.in > $TEST/Makefile
rm -f $TMP
(cd $TEST; make clean > /dev/null 2> /dev/null)

# check: ranlib command is available or not
$ECHON "checking: ranlib command is available... \c"
TMP=`searchpath ranlib $BINPATH`
if [ "$TMP" = "" ]; then
    RANLIB="no"
fi
echo $RANLIB

# check: calling C function from fortran
$ECHON "checking: call C function from fortran... \c"
(cd $TEST; make a1.out > /dev/null 2> /dev/null) && CTOF="no need"
echo "$CTOF '_'"
if [ "$CTOF" = "no need" ]; then
    if [ ! -f $LIB/flib.c.org ]; then
        mv -f $LIB/flib.c $LIB/flib.c.org
    fi
    sed "s/_(/(/" $LIB/flib.c.org > $LIB/flib.c
fi

# check: select system call
$ECHON "checking: call select() system call... \c"
(cd $TEST; make a2.out > /dev/null 2> /dev/null) || USE_SELECT="failed"
echo $USE_SELECT

# check: opendir system call
$ECHON "checking: call opendir() system call... \c"
(cd $TEST; make a3.out > /dev/null 2> /dev/null) || USE_OPENDIR="failed"
echo $USE_OPENDIR

# check: scandir system call
$ECHON "checking: call scandir() system call... \c"
(cd $TEST; make a4.out > /dev/null 2> /dev/null) || USE_SCANDIR="failed"
echo $USE_SCANDIR

if [ $USE_LIB_XPM = "yes" ]; then
    echo "use xpm library."
fi
if [ $USE_LIB_UNGIF = "yes" ]; then
#    LIBUNGIF_PATH="/usr/local/lib"
    LIBUNGIF_PATH=`searchpath libungif.a $LIBPATH`
    while [ "$LIBUNGIF_PATH" = "" ]; do
        ask_str "Directory where libungif.a installed []: "
        if [ "$TMP" != "" ]; then
            LIBUNGIF_PATH="$TMP"
        fi
    done
    LIBUNGIF_IPATH=`echo $LIBUNGIF_PATH | sed 's/lib/include/'`
    echo "use Libungif library: $LIBUNGIF_PATH/libungif.a"
elif [ $USE_LIB_GIF = "yes" ]; then
#    LIBGIF_PATH="/usr/local/lib"
    LIBGIF_PATH=`searchpath libgif.a $LIBPATH`
    while [ "$LIBGIF_PATH" = "" ]; do
        ask_str "directory where libgif.a installed []: "
        if [ "$TMP" != "" ]; then
            LIBGIF_PATH="$TMP"
        fi
    done
    LIBGIF_IPATH=`echo $LIBGIF_PATH | sed 's/lib/include/'`
    echo "use GifLib library: $LIBGIF_PATH/libgif.a"
fi

# generate "Makefile" from "Makefile.in"
TMP=sedscript_$$
cat >$TMP << _EOF_
s!XROOT = .*!XROOT = $XROOT!
s!CC = .*!CC = $CC!
s!prefix = .*!prefix = $INSTDIR!
_EOF_
if [ "$RANLIB" = "no" ]; then
    echo 's!	ranlib!#	ranlib!' >> $TMP
fi
if [ "$USE_LIB_XPM" = "yes" ]; then
    echo 's!#LIBXPM!LIBXPM!' >> $TMP
fi
if [ "$USE_LIB_UNGIF" = "yes" ]; then
    echo "s!#LIBUNGIF.*!LIBUNGIF = -L$LIBUNGIF_PATH -lungif!" >> $TMP
    echo "s!#UNGIF_IPATH.*!UNGIF_IPATH = -I$LIBUNGIF_IPATH!" >> $TMP
elif [ "$USE_LIB_GIF" = "yes" ]; then
    echo "s!#LIBGIF.*!LIBGIF = -L$LIBGIF_PATH -lgif!" >> $TMP
    echo "s!#GIF_IPATH.*!GIF_IPATH = -I$LIBGIF_IPATH!" >> $TMP
fi

echo "generating \"Makefile\"..."
for i in . $PGW $LIB $PTR $TEST; do
    if [ -f $i/Makefile ]; then
        mv -f $i/Makefile $i/Makefile.bac
    fi
    sed -f $TMP $i/Makefile.in > $i/Makefile
done
rm -f $TMP


# generate "config.h" from "config.h.in"
TMP=sedscript_$$
if [ $OS = "BSD" ]; then
    echo 's!.*#undef PTR_BSD.*!#define PTR_BSD!' >> $TMP
    echo 's!.*#define PTR_SYSV.*!#undef PTR_SYSV!' >> $TMP
else
    echo 's!.*#undef PTR_SYSV.*!#define PTR_SYSV!' >> $TMP
    echo 's!.*#define PTR_BSD.*!#undef PTR_BSD!' >> $TMP
fi
if [ $USE_SELECT = "failed" ]; then
    echo 's!.*#undef USE_SELECT.*!#undef USE_SELECT!' >> $TMP
fi
if [ $USE_OPENDIR = "failed" ]; then
    echo 's!.*#define NO_OPENDIR.*!#define NO_OPENDIR!' >> $TMP
fi
if [ $USE_SCANDIR = "failed" ]; then
    echo 's!.*#define NO_SCANDIR.*!#define NO_SCANDIR!' >> $TMP
fi
if [ $USE_LIB_XPM = "yes" ]; then
    echo 's!.*#define USE_LIB_XPM.*!#define USE_LIB_XPM!' >> $TMP
fi
if [ $USE_LIB_UNGIF = "yes" ]; then
    echo 's!.*#define USE_LIB_UNGIF.*!#define USE_LIB_UNGIF!' >> $TMP
elif [ $USE_LIB_GIF = "yes" ]; then
    echo 's!.*#define USE_LIB_GIF.*!#define USE_LIB_GIF!' >> $TMP
fi
if [ -f ./config.h ]; then
    mv -f ./config.h ./config.h.bac
fi

echo "generating \"config.h\"..."
sed -f $TMP ./config.h.in > ./config.h
rm -f $TMP


cat > config.log << _EOF_
XROOT="$XROOT"
CC="$CC"
INSTDIR="$INSTDIR"
OS="$OS"
RANLIB="$RANLIB"
CTOF="$CTOF"
USE_SELECT="$USE_SELECT"
USE_OPENDIR="$USE_OPENDIR"
USE_SCANDIR="$USE_SCANDIR"
USE_LIB_XPM="$USE_LIB_XPM"
#USE_LIB_UNGIF="$USE_LIB_UNGIF"
#USE_LIB_GIF="$USE_LIB_GIF"
_EOF_
