#include "ptr.h"
#include <stdlib.h>



void FreeData(PTRData* dp){
    if(dp->x1) free(dp->x1);
    if(dp->x2) free(dp->x2);
    if(dp->x3) free(dp->x3);
    if(dp->v){
        int i;
        for(i=0; i<dp->nv; i++) free(dp->v[i]);
        free(dp->v);
    }
    free(dp);
}

void FreePoints(PTRPoints* gp){
    FreeData(gp->dp);
    free(gp);
}
void FreeLine(PTRLine* gp){
    FreeData(gp->dp);
    free(gp);
}
#ifdef DEBUG
void FreeMeshLines(PTRMeshLines* gp){
    FreeData(gp->dp);
    free(gp);
}
#endif
void FreeString(PTRString* gp){
    if(gp->str) free(gp->str);
    free(gp);
}
void FreeFixString(PTRFixString* gp){
    if(gp->str) free(gp->str);
    free(gp);
}
void FreeGraphSet(PTRGraphSet* gp)
{
    PTRGraphList* gpl;
    gpl=gp->gpl.next;
    while(gpl){
        FreeGraph(gpl->gp);
        gpl=gpl->next;
    }
    FreeGraphLists(gp->gpl.next);
    free(gp);
}
void FreeScale(PTRScale* gp){
    if(gp->x1) free(gp->x1);
    if(gp->x2) free(gp->x2);
    if(gp->x3) free(gp->x3);
    free(gp);
}
void FreeSclFrame(PTRSclFrame* gp){
    free(gp);
}


void FreeGraph(PTRGraph* gp)
{
    switch(gp->type){
      case POINTS:
        FreePoints((PTRPoints*)gp);
        break;
      case LINE:
        FreeLine((PTRLine*)gp);
        break;
      case STRING:
        FreeString((PTRString*)gp);
        break;
      case FIXSTRING:
        FreeFixString((PTRFixString*)gp);
        break;
      case GRAPHSET:
        FreeGraphSet((PTRGraphSet*)gp);
        break;
      case LABEL:
        FreeString((PTRString*)gp);
        break;
      case SCALE:
        FreeScale((PTRScale*)gp);
        break;
      case SCLFRAME:
        FreeSclFrame((PTRSclFrame*)gp);
        break;
    }
}

void FreeGraphList(PTRGraphList* gpl)
{
    free(gpl);
}
void FreeGraphLists(PTRGraphList* gpl)
{
    PTRGraphList* gpl1;
    while(gpl){
        gpl1=gpl->next;
        FreeGraphList(gpl);
        gpl=gpl1;
    }
}
