#include "ptr.h"
#include <stdlib.h>
#include <math.h>


PGObjectList* PointsToPGObjects(PTRWindow* wp, PTRPoints* gp)
{
    PGObject* pg;
    PGObjectList* pgl;
    pgl=(PGObjectList*)malloc(sizeof(PGObjectList));
    pgl->next=NULL;
    pg=(PGObject*)CreatePGPoints(gp->dp->nx1, gp->x1, gp->x2,
                                 gp->x3, gp->pgp);
    PushToPGObjectList(pgl, pg);
    return pgl;
}
PGObjectList* LineToPGObjects(PTRWindow* wp, PTRLine* gp)
{
    PGObject* pg;
    PGObjectList* pgl;
    pgl=(PGObjectList*)malloc(sizeof(PGObjectList));
    pgl->next=NULL;
    pg=(PGObject*)CreatePGLine(gp->dp->nx1, gp->x1, gp->x2,
                               gp->x3, gp->pgp);
    PushToPGObjectList(pgl, pg);
    return pgl;
}
PGObjectList* StringToPGObjects(PTRWindow* wp, PTRString* gp)
{
    PGObject* pg;
    PGObjectList* pgl;
    pgl=(PGObjectList*)malloc(sizeof(PGObjectList));
    pgl->next=NULL;
    pg=(PGObject*)CreatePGString(gp->x1, gp->x2, gp->x3,
                                 gp->str, gp->pgp);
    PushToPGObjectList(pgl, pg);
    return pgl;
}
PGObjectList* FixStringToPGObjects(PTRWindow* wp, PTRFixString* gp)
{
    PGObject* pg;
    PGObjectList* pgl;
    pgl=(PGObjectList*)malloc(sizeof(PGObjectList));
    pgl->next=NULL;
    pg=(PGObject*)CreatePGFixString(gp->x, gp->y, gp->str, gp->pgp);
    PushToPGObjectList(pgl, pg);
    return pgl;
}
#ifdef DEBUG
PGObjectList* ArrowsToPGObjects(PTRWindow* wp, PTRArrows* gp)
{
    double* x1;  double* x2;  double* x3;
    double* v1;  double* v2;  double* v3;
    PGObject* pg;
    PGObjectList* pgl;
    switch(wp->dim){
      case D1_1:
      case D2_1:
        return NULL;
        break;
      case D3:
        if(gp->dp->dim<3){
/*            SetError(WARN, "ArrowsToPGObjects", E_NoEnoughData);*/
            return NULL;
        }
        break;
      case D2:
        if(gp->dp->dim<2){
/*            SetError(WARN, "ArrowsToPGObjects", E_NoEnoughData);*/
            return NULL;
        }
        break;
    }
    pgl=(PGObjectList*)malloc(sizeof(PGObjectList));
    pgl->next=NULL;
    switch(wp->dim){
      case D2:
        x1=gp->dp->x1;  x2=gp->dp->x2;  x3=NULL;
        v1=gp->dp->v[0];
        v2=gp->dp->v[1];  v3=NULL;
        break;
      case D3:
        x1=gp->dp->x1;  x2=gp->dp->x2;  x3=gp->dp->x3;
        v1=gp->dp->v[0];
        v2=gp->dp->v[1];
        v3=gp->dp->v[2];
        break;
    }
    pg=(PGObject*)CreatePGArrows(gp->dp->nx1, x1, x2, x3,
                                 v1, v2, v3, gp->exp, gp->pgp);
    PushToPGObjectList(pgl, pg);
    return pgl;
}
#endif

PGObjectList* ScaleToPGObjects(PTRWindow* wp, PTRScale* gp)
{
    int i, j, ie;
    double x3=0.0;
    double* tx3=NULL;
    PGObject* pg;
    PGObjectList* pgl;
    pgl=(PGObjectList*)malloc(sizeof(PGObjectList));
    pgl->next=NULL;
    for(ie=0; ie<(gp->nx)/(gp->num); ie++){
    for(i=0; i<gp->num; i++){
        j=5*(i+ie*gp->num);
        if(wp->ax3) tx3=gp->x3+j;
        pg=(PGObject*)CreatePGLine(2, gp->x1+j, gp->x2+j, tx3,
                                   gp->pgp);
        PushToPGObjectList(pgl, pg);
        if(wp->ax3) tx3=gp->x3+j+2;
        pg=(PGObject*)CreatePGLine(2, gp->x1+j+2, gp->x2+j+2, tx3,
                                   gp->pgp);
        PushToPGObjectList(pgl, pg);
        if(gp->scl[i].lb[0]=='\0') continue;
        if(wp->coord==CYL) continue;
        if(gp->x3) x3=gp->x3[j+4];
        pg=(PGObject*)
                CreatePGString(gp->x1[j+4], gp->x2[j+4], x3,
                               gp->scl[i].lb, gp->pgp);
        PushToPGObjectList(pgl, pg);
    }}
    return pgl;
}
PGObjectList* SclFrameToPGObjects(PTRWindow* wp, PTRSclFrame* gp)
{
    PGObject* pg;
    PGObjectList* pgl;
    pgl=(PGObjectList*)malloc(sizeof(PGObjectList));
    pgl->next=NULL;
    pg=(PGObject*)CreatePGLine(gp->num, gp->x1, gp->x2, gp->x3,
                               gp->pgp);
    PushToPGObjectList(pgl, pg);
    return pgl;
}


PGObjectList* ToPGObjects(PTRWindow* wp, PTRGraph* gp)
{
    PGObjectList* pgl;
    if(gp->any.vis==FALSE) return NULL;
    EnterFunc("ToPGObjects");
    if(gp->type==POINTS){
        pgl=PointsToPGObjects(wp, (PTRPoints*)gp);
    }else if(gp->type==LINE){
        pgl=LineToPGObjects(wp, (PTRLine*)gp);
    }else if(gp->type==STRING){
        pgl=StringToPGObjects(wp, (PTRString*)gp);
    }else if(gp->type==FIXSTRING){
        pgl=FixStringToPGObjects(wp, (PTRFixString*)gp);
/*    }else if(gp->type==ARROWS){
        pgl=ArrowsToPGObjects(wp, (PTRArrows*)gp);*/
    }else if(gp->type==GRAPHSET){
    }else if(gp->type==LABEL){
        pgl=StringToPGObjects(wp, (PTRString*)gp);
    }else if(gp->type==SCALE){
        pgl=ScaleToPGObjects(wp, (PTRScale*)gp);
    }else if(gp->type==SCLFRAME){
        pgl=SclFrameToPGObjects(wp, (PTRSclFrame*)gp);
    }else{
        SetError(INNER, "Don't know how to make the graph.\n");
        LeaveFunc();  return NULL;
    }
    LeaveFunc();  return pgl;
}



/* clear graph */

PTRGraphList* SearchGraph(PTRGraphList* top, PTRGraph* gp)
{
    PTRGraphList* gpl;
    gpl=top;
    while(gpl->next && gpl->next->gp != gp){
        gpl=gpl->next;
    }
    if(gpl->next){
        return gpl;
    }else{
        return NULL;
    }
}
void RemovePGObjectsOf(PTRWindow* wp, PTRGraph* gp)
{
    if(gp->type==GRAPHSET){
        PTRGraphList* gpl;
        gpl=gp->graphset.gpl.next;
        while(gpl){
            RemovePGObjectsOf(wp, gpl->gp);
            gpl=gpl->next;
        }
    }else{
        RemovePGObjects(wp->pgw, gp->any.pgl);
        gp->any.pgl=NULL;
    }
}
void RemoveGraph(PTRWindow* wp, PTRGraph* gp)
{
    int i;
    PTRGraphList* gpl;
    PTRGraphList* prev;
    EnterFunc("RemoveGraph");
    if(!(prev=SearchGraph(&(wp->gpl), gp))){
        SetError(ERROR, E_NoSuchGraph);
        LeaveFunc();  return;
    }
    gpl=prev->next;
    RemovePGObjectsOf(wp, gp);
    prev->next=gpl->next;
    if(gp->type==LABEL){
        for(i=0; i<5; i++){
            if(wp->label[i]==(PTRLabel*)gp) wp->label[i]=NULL;
        }
    }else if(gp->type==SCALE){
        for(i=1; i<5; i++){
            if(wp->scale[i]==(PTRScale*)gp) wp->scale[i]=NULL;
        }
    }else if(gp==(PTRGraph*)(wp->frame)){
        wp->frame=NULL;
    }
    FreeGraph(gp);
    FreeGraphList(gpl);
    LeaveFunc();
}

void ClearWindow(PTRWindow* wp)
{
    PTRGraphList* gpl;
    PTRGraphList* prev;
    prev=&(wp->gpl);
    gpl=prev->next;
    while(gpl){
        if(gpl->gp->type!=SCALE
           && gpl->gp->type!=SCLFRAME
           && gpl->gp->type!=LABEL){
            RemovePGObjectsOf(wp, gpl->gp);
            FreeGraph(gpl->gp);
            prev->next=gpl->next;
            FreeGraphList(gpl);
        }else{
            prev=prev->next;
        }
        gpl=prev->next;
    }
}
void ClearWindows(PTRWindow* wp)
{
    PTRWindowList* wpl;
    wpl=wp->wpl.next;
    while(wpl){
	ClearWindows(wpl->wp);
	wpl=wpl->next;
    }
    ClearWindow(wp);
}


/* update rgaphs */
void UpdatePTRPoints(PTRWindow* wp, PTRPoints* gp)
{
    gp->vis=FALSE;
    gp->x1=NULL;  gp->x2=NULL;  gp->x3=NULL;
    if(wp->ax1==AX_X1){
        gp->x1=gp->dp->x1;
    }else if(wp->ax1==AX_X2){
        gp->x1=gp->dp->x2;
    }else if(wp->ax1==AX_X3){
        gp->x1=gp->dp->x3;
    }else if(wp->ax1==AX_V){
        gp->x1=gp->dp->v[0];
    }
    if(wp->ax2==AX_X1){
        gp->x2=gp->dp->x1;
    }else if(wp->ax2==AX_X2){
        gp->x2=gp->dp->x2;
    }else if(wp->ax2==AX_X3){
        gp->x2=gp->dp->x3;
    }else if(wp->ax2==AX_V){
        gp->x2=gp->dp->v[0];
    }
    if(wp->ax3==AX_X1){
        gp->x3=gp->dp->x1;
    }else if(wp->ax3==AX_X2){
        gp->x3=gp->dp->x2;
    }else if(wp->ax3==AX_X3){
        gp->x3=gp->dp->x3;
    }else if(wp->ax3==AX_V){
        gp->x3=gp->dp->v[0];
    }
    if(gp->x1==NULL || gp->x2==NULL){
        gp->vis=FALSE;
    }else if(wp->ax3 && gp->x3==NULL){
        gp->vis=FALSE;
    }else{
        gp->vis=TRUE;
    }
}
void UpdatePTRLine(PTRWindow* wp, PTRLine* gp)
{
    gp->vis=FALSE;
    gp->x1=NULL;  gp->x2=NULL;  gp->x3=NULL;
    if(wp->ax1==AX_X1){
        gp->x1=gp->dp->x1;
    }else if(wp->ax1==AX_X2){
        gp->x1=gp->dp->x2;
    }else if(wp->ax1==AX_X3){
        gp->x1=gp->dp->x3;
    }else if(wp->ax1==AX_V){
        gp->x1=gp->dp->v[0];
    }
    if(wp->ax2==AX_X1){
        gp->x2=gp->dp->x1;
    }else if(wp->ax2==AX_X2){
        gp->x2=gp->dp->x2;
    }else if(wp->ax2==AX_X3){
        gp->x2=gp->dp->x3;
    }else if(wp->ax2==AX_V){
        gp->x2=gp->dp->v[0];
    }
    if(wp->ax3==AX_X1){
        gp->x3=gp->dp->x1;
    }else if(wp->ax3==AX_X2){
        gp->x3=gp->dp->x2;
    }else if(wp->ax3==AX_X3){
        gp->x3=gp->dp->x3;
    }else if(wp->ax3==AX_V){
        gp->x3=gp->dp->v[0];
    }
    if(gp->x1==NULL || gp->x2==NULL){
        gp->vis=FALSE;
    }else if(wp->ax3 && gp->x3==NULL){
        gp->vis=FALSE;
    }else{
        gp->vis=TRUE;
    }
}
void UpdatePTRString(PTRWindow* wp, PTRString* gp)
{
    gp->x1=gp->x[wp->ax1];
    gp->x2=gp->x[wp->ax2];
    if(wp->ax3) gp->x3=gp->x[wp->ax3];
    gp->vis=TRUE;
}
void UpdatePTRLabel(PTRWindow* wp, PTRLabel* gp)
{
    if(gp==NULL) return;
    gp->vis=FALSE;
    if(gp!=wp->label[gp->ax]) return;
    if(gp->ax==AX_NL){
        gp->x1=0.5*(wp->min[wp->ax1]+wp->max[wp->ax1]);
        if(wp->ax3){
            gp->x2=0.5*(wp->min[wp->ax2]+wp->max[wp->ax2]);
            gp->x3=wp->max[wp->ax3]+wp->size[wp->ax3]*gp->lb_dist;
        }else{
            gp->x2=wp->max[wp->ax2]
                    +wp->size[wp->ax2]*0.5*gp->lb_dist;
        }
    }else if(wp->ax1==gp->ax){
        gp->x1=0.5*(wp->min[wp->ax1]+wp->max[wp->ax1]);
        if(wp->ax3){
            gp->x3=wp->min[wp->ax3];
            gp->x2=wp->min[wp->ax2]
                    -wp->size[wp->ax2]*2.0*gp->lb_dist;
        }else{
            gp->x2=wp->min[wp->ax2]-wp->size[wp->ax2]*gp->lb_dist;
        }
    }else if(wp->ax2==gp->ax){
        gp->x2=0.5*(wp->min[wp->ax2]+wp->max[wp->ax2]);
        if(wp->ax3){
            gp->x3=wp->min[wp->ax3];
            gp->x1=wp->max[wp->ax1]
                    +wp->size[wp->ax1]*2.0*gp->lb_dist;
        }else{
            gp->x1=wp->min[wp->ax1]-wp->size[wp->ax1]*gp->lb_dist;
        }
    }else if(wp->ax3==gp->ax){
        gp->x1=wp->max[wp->ax1]+1.5*wp->size[wp->ax1]*gp->lb_dist;
        gp->x2=wp->max[wp->ax2]+1.5*wp->size[wp->ax2]*gp->lb_dist;
        gp->x3=0.5*(wp->min[wp->ax3]+wp->max[wp->ax3]);
    }
    gp->vis=TRUE;
}
void UpdatePTRScale(PTRWindow* wp, PTRScale* gp)
{
    int i, j, ie, nx=0, size;
    double c=1.0;
    double x1min, x1max, x1size, x2min, x2max, x2size;
    double x3min, x3max, x3size;
    if(gp==NULL) return;
    gp->vis=FALSE;
    if(gp!=wp->scale[gp->ax]) return;
    if(wp->coord==CYL) return;
    if(wp->ax1==gp->ax) nx+=gp->num;
    if(wp->ax2==gp->ax) nx+=gp->num;
    if(wp->ax3==gp->ax) nx+=gp->num;
    if(nx==0) return;
    size=5*nx*sizeof(double);
    if(gp->x1){
        gp->x1=(double*)realloc(gp->x1, size);
    }else{
        gp->x1=(double*)malloc(size);
    }
    if(gp->x2){
        gp->x2=(double*)realloc(gp->x2, size);
    }else{
        gp->x2=(double*)malloc(size);
    }
    if(wp->ax3){
        c=2.0*c;
        if(gp->x3){
            gp->x3=(double*)realloc(gp->x3, size);
        }else{
            gp->x3=(double*)malloc(size);
        }
    }
    gp->nx=nx;
    ie=0;
    if(wp->ax1==gp->ax){
        x2min=wp->min[wp->ax2];  x2max=wp->max[wp->ax2];
        x2size=x2max-x2min;
        if(wp->ax3){
            x3min=wp->min[wp->ax3];  x3max=wp->max[wp->ax3];
            x3size=x3max-x3min;
        }
        for(i=0; i<gp->num; i++){
            j=5*(i+ie*gp->num);
            gp->x1[j]=gp->scl[i].x;
            gp->x2[j]=x2min;
            if(wp->ax3) gp->x3[j]=x3min;
            gp->x1[j+1]=gp->scl[i].x;
            if(gp->scl[i].lb[0]!='\0'){
                gp->x2[j+1]=x2min+gp->scl_dist*x2size;
            }else{
                gp->x2[j+1]=x2min+gp->sc_dist*x2size;
            }
            if(wp->ax3) gp->x3[j+1]=x3min;
            gp->x1[j+2]=gp->scl[i].x;
            gp->x2[j+2]=x2max;
            if(wp->ax3) gp->x3[j+2]=x3min;
            gp->x1[j+3]=gp->scl[i].x;
            if(gp->scl[i].lb[0]!='\0'){
                gp->x2[j+3]=x2max-gp->scl_dist*x2size;
                gp->x1[j+4]=gp->scl[i].x;
                gp->x2[j+4]=x2min-c*gp->lb_dist*x2size;
                if(gp->x3) gp->x3[j+4]=x3min;
            }else{
                gp->x2[j+3]=x2max-gp->sc_dist*x2size;
            }
            if(wp->ax3) gp->x3[j+3]=x3min;
        }
        ie++;
    }
    if(wp->ax2==gp->ax){
        x1min=wp->min[wp->ax1];  x1max=wp->max[wp->ax1];
        x1size=x1max-x1min;
        if(wp->ax3){
            x3min=wp->min[wp->ax3];  x3max=wp->max[wp->ax3];
            x3size=x3max-x3min;
        }
        for(i=0; i<gp->num; i++){
            j=5*(i+ie*gp->num);
            gp->x1[j]=x1min;
            gp->x2[j]=gp->scl[i].x;
            if(wp->ax3) gp->x3[j]=x3min;
            gp->x2[j+1]=gp->scl[i].x;
            if(gp->scl[i].lb[0]!='\0'){
                gp->x1[j+1]=x1min+gp->scl_dist*x1size;
            }else{
                gp->x1[j+1]=x1min+gp->sc_dist*x1size;
            }
            if(wp->ax3) gp->x3[j+1]=x3min;
            gp->x1[j+2]=x1max;
            gp->x2[j+2]=gp->scl[i].x;
            if(wp->ax3) gp->x3[j+2]=x3min;
            gp->x2[j+3]=gp->scl[i].x;
            if(gp->scl[i].lb[0]!='\0'){
                gp->x1[j+3]=x1max-gp->scl_dist*x1size;
                gp->x2[j+4]=gp->scl[i].x;
                if(wp->ax3){
                    gp->x1[j+4]=x1max+c*gp->lb_dist*x1size;
                    gp->x3[j+4]=x3min;
                }else{
                    gp->x1[j+4]=x1min-c*gp->lb_dist*x1size;
                }
            }else{
                gp->x1[j+3]=x1max-gp->sc_dist*x1size;
            }
            if(wp->ax3) gp->x3[j+3]=x3min;
        }
        ie++;
    }
    if(wp->ax3==gp->ax){
        x1min=wp->min[wp->ax1];  x1max=wp->max[wp->ax1];
        x1size=x1max-x1min;
        x2min=wp->min[wp->ax2];  x2max=wp->max[wp->ax2];
        x2size=x2max-x2min;
        for(i=0; i<gp->num; i++){
            j=5*(i+ie*gp->num);
            gp->x1[j]=x1min;
            gp->x2[j]=x2max;
            gp->x3[j]=gp->scl[i].x;
            gp->x2[j+1]=x2max;
            gp->x3[j+1]=gp->scl[i].x;
            if(gp->scl[i].lb[0]!='\0'){
                gp->x1[j+1]=x1min+gp->scl_dist*x1size;
            }else{
                gp->x1[j+1]=x1min+gp->sc_dist*x1size;
            }
            gp->x1[j+2]=x1max;
            gp->x2[j+2]=x2max;
            gp->x3[j+2]=gp->scl[i].x;
            gp->x2[j+3]=x2max;
            gp->x3[j+3]=gp->scl[i].x;
            if(gp->scl[i].lb[0]!='\0'){
                gp->x1[j+3]=x1max-gp->scl_dist*x1size;
                gp->x1[j+4]=x1max+c*gp->lb_dist*x1size;
                gp->x2[j+4]=x2max;
                gp->x3[j+4]=gp->scl[i].x;
            }else{
                gp->x1[j+3]=x1max-gp->sc_dist*x1size;
            }
        }
    }
    gp->vis=TRUE;
}
void UpdatePTRSclFrame(PTRWindow* wp, PTRSclFrame* gp)
{
    double x1min, x1max, x2min, x2max, x3min, x3max;
    if(gp==NULL) return;
    gp->vis=FALSE;
    if(gp!=wp->frame) return;
    x1min=wp->min[wp->ax1];  x1max=wp->max[wp->ax1];
    x2min=wp->min[wp->ax2];  x2max=wp->max[wp->ax2];
    x3min=wp->min[wp->ax3];  x3max=wp->max[wp->ax3];
    if(wp->coord==CAR){
        gp->num=5;
        gp->x1[0]=x1min;  gp->x2[0]=x2min;  gp->x3[0]=x3min;
        gp->x1[1]=x1max;  gp->x2[1]=x2min;  gp->x3[1]=x3min;
        gp->x1[2]=x1max;  gp->x2[2]=x2max;  gp->x3[2]=x3min;
        gp->x1[3]=x1min;  gp->x2[3]=x2max;  gp->x3[3]=x3min;
        gp->x1[4]=x1min;  gp->x2[4]=x2min;  gp->x3[4]=x3min;
        if(wp->ax3){
            gp->num=11;
            gp->x1[5]=x1min;  gp->x2[5]=x2min;  gp->x3[5]=x3max;
            gp->x1[6]=x1min;  gp->x2[6]=x2max;  gp->x3[6]=x3max;
            gp->x1[7]=x1min;  gp->x2[7]=x2max;  gp->x3[7]=x3min;
            gp->x1[8]=x1max;  gp->x2[8]=x2max;  gp->x3[8]=x3min;
            gp->x1[9]=x1max;  gp->x2[9]=x2max;  gp->x3[9]=x3max;
            gp->x1[10]=x1min; gp->x2[10]=x2max; gp->x3[10]=x3max;
        }
    }else if(wp->coord==CYL){
        int i, i1, i2, i3, i4, nn=33;
        double x2;
        gp->num=nn;
        for(i=0; i<nn; i++){
            i1=i;  i2=i+nn;
            x2=x2min+i*(x2max-x2min)/(nn-1);
            gp->x1[i1]=x1max;  gp->x2[i1]=x2;  gp->x3[i1]=x3min;
            gp->x1[i2]=x1max;  gp->x2[i2]=x2;  gp->x3[i2]=x3max;
        }
        if(wp->ax3){
            i=2*nn;
            gp->num=2*nn+3;
            gp->x1[i]=x1min;  gp->x2[i]=x2min;  gp->x3[i]=x3max;
            i++;
            gp->x1[i]=x1min;  gp->x2[i]=x2min;  gp->x3[i]=x3min;
            i++;
            gp->x1[i]=x1max;  gp->x2[i]=x2min;  gp->x3[i]=x3min;
        }
    }else{
        return;
    }
    gp->vis=TRUE;
}
void UpdatePTRGProperties(PTRWindow* wp, PTRGraph* gp)
{
    if(gp==NULL) return;
    switch(gp->type){
      case POINTS:
        UpdatePTRPoints(wp, (PTRPoints*)gp);
        break;
      case LINE:
        UpdatePTRLine(wp, (PTRLine*)gp);
        break;
      case STRING:
        UpdatePTRString(wp, (PTRString*)gp);
        break;
      case LABEL:
        UpdatePTRLabel(wp, (PTRLabel*)gp);
        break;
      case SCALE:
        UpdatePTRScale(wp, (PTRScale*)gp);
        break;
      case SCLFRAME:
        UpdatePTRSclFrame(wp, (PTRSclFrame*)gp);
        break;
    }
}

/* draw graphs */
void ToAndAddPGObjects(PTRWindow* wp, PTRGraph* gp)
{
    if(gp->type==GRAPHSET){
        PTRGraphList* gpl;
        gpl=gp->graphset.gpl.next;
        while(gpl){
            ToAndAddPGObjects(wp, gpl->gp);
            gpl=gpl->next;
        }
    }else{
        UpdatePTRGProperties(wp, gp);
        gp->any.pgl=ToPGObjects(wp, gp);
        AddPGObjects(wp->pgw, gp->any.pgl);
    }
}

void AddPTRGraph(PTRWindow* wp, PTRGraph* gp)
{
    PTRGraphList* gpl;
    gpl=(PTRGraphList*)malloc(sizeof(PTRGraphList));
    gpl->gp=gp;
    gpl->next=wp->gpl.next;
    wp->gpl.next=gpl;
    ToAndAddPGObjects(wp, gp);
}
void UpdatePTRGraph(PTRWindow* wp, PTRGraph* gp)
{
    if(gp==NULL) return;
    RemovePGObjectsOf(wp, gp);
    ToAndAddPGObjects(wp, gp);
}
void UpdatePTRGraphs(PTRWindow* wp)
{
    PTRGraphList* gpl;
    gpl=wp->gpl.next;
    while(gpl){
        UpdatePTRGraph(wp, gpl->gp);
        gpl=gpl->next;
    }
}

void AddToGraphSet(PTRGraphSet* set, PTRGraph* gp)
{
    PTRGraphList* gpl;
    if(gp==NULL) return;
    gpl=(PTRGraphList*)malloc(sizeof(PTRGraphList));
    gpl->gp=gp;
    gpl->next=set->gpl.next;
    set->gpl.next=gpl;
}
PTRGraphSet* CreateGraphSet()
{
    PTRGraphSet* gp;
    gp=(PTRGraphSet*)malloc(sizeof(PTRGraphSet));
    gp->type=GRAPHSET;
    gp->vis=TRUE;
    gp->pgl=NULL;
    gp->gpl.next=NULL;
    return gp;
}
PTRGraph* CreatePoints(PTRData* data, int l, PGP pgp)
{
    PTRData* dp;
    PTRPoints* gp;
    EnterFunc("CreatePoints");
    if(data){
        dp=data;
    }else if(!(dp=CurrentData())){
        SetError(ERROR, E_NoData);
        LeaveFunc();  return NULL;
    }
    if(dp->type!=SEQ){
        SetError(ERROR, E_BadDataType);
        LeaveFunc();  return NULL;
    }else if(dp->nv<l){
        SetError(ERROR, E_NoEnoughLeng);
        LeaveFunc();  return NULL;
    }
    gp=(PTRPoints*)malloc(sizeof(PTRPoints));
    gp->type=POINTS;
    gp->vis=TRUE;
    if(0<l){
        gp->dp=CreateSeqData(dp->dim, dp->nx1, 1, dp->x1,
                             dp->x2, dp->x3, dp->v+l-1);
    }else{
        gp->dp=CreateSeqData(dp->dim, dp->nx1, 0, dp->x1,
                              dp->x2, dp->x3, NULL);
    }
    gp->pgp=pgp;
    LeaveFunc();  return (PTRGraph*)gp;
}
PTRGraph* DrawPoints(PTRWindow* wp, PTRData* data, int l, PGP pgp)
{
    PTRGraph* gp;
    EnterFunc("DrawPoints");
    if((gp=CreatePoints(data, l, pgp))) AddPTRGraph(wp, gp);
    LeaveFunc();  return gp;
}
PTRGraph* CreateLine(PTRData* data, int l, PGP pgp)
{
    PTRData* dp;
    PTRLine* gp;
    EnterFunc("CreateLine");
    if(data){
        dp=data;
    }else if(!(dp=CurrentData())){
        SetError(ERROR, E_NoData);
        LeaveFunc();  return NULL;
    }
    if(dp->type != SEQ){
        SetError(ERROR, E_BadDataType);
        LeaveFunc();  return NULL;
    }else if(dp->nv<l){
        SetError(ERROR, E_NoEnoughLeng);
        LeaveFunc();  return NULL;
    }
    gp=(PTRLine*)malloc(sizeof(PTRLine));
    gp->type=LINE;
    gp->vis=TRUE;
    if(0<l){
        gp->dp=CreateSeqData(dp->dim, dp->nx1, 1, dp->x1,
                             dp->x2, dp->x3, dp->v+l-1);
    }else{
        gp->dp=CreateSeqData(dp->dim, dp->nx1, 0, dp->x1,
                              dp->x2, dp->x3, NULL);
    }
    gp->pgp=pgp;
    LeaveFunc();  return (PTRGraph*)gp;
}
PTRGraph* DrawLine(PTRWindow* wp, PTRData* data, int l, PGP pgp)
{
    PTRGraph* gp;
    EnterFunc("DrawLine");
    if((gp=CreateLine(data, l, pgp))) AddPTRGraph(wp, gp);
    LeaveFunc();  return gp;
}
PTRGraph* CreateMeshLines(PTRData* data, int l, PGP pgp)
{
    int i, j, k, nk=1, num=20;
    PTRData* dp;
    PTRData* dp1;
    PTRGraphSet* gp;
    EnterFunc("CreateMeshLines");
    if(data){
        dp=data;
    }else if(!(dp=CurrentData())){
        SetError(ERROR, E_NoData);
        LeaveFunc();  return NULL;
    }
    if(l<=0 || dp->nv<l){
        SetError(ERROR, E_BadValue);
        LeaveFunc();  return NULL;
    }
    if(dp->type!=GRID){
        SetError(ERROR, E_BadDataType);
        LeaveFunc();  return NULL;
    }else if(dp->dim<2){
        SetError(ERROR, E_NoEnoughDim);
        LeaveFunc();  return NULL;
    }
    gp=CreateGraphSet();
    if(dp->dim==3) nk=dp->nx3;
    for(k=0; k<nk; k+=nk/num+1){
    for(j=0; j<dp->nx2; j+=dp->nx2/num+1){
        dp1=ExtractSeqDataI(dp, j, k);
        if(dp1){
            AddToGraphSet(gp, CreateLine(dp1, l, pgp));
            FreeData(dp1);
        }
    }}
    for(k=0; k<nk; k+=nk/num+1){
    for(i=0; i<dp->nx1; i+=dp->nx1/num+1){
        dp1=ExtractSeqDataJ(dp, i, k);
        if(dp1){
            AddToGraphSet(gp, CreateLine(dp1, l, pgp));
            FreeData(dp1);
        }
    }}
    if(dp->dim==3){
    for(j=0; j<dp->nx2; j+=dp->nx2/num+1){
    for(i=0; i<dp->nx1; i+=dp->nx1/num+1){
        dp1=ExtractSeqDataK(dp, i, j);
        if(dp1){
            AddToGraphSet(gp, CreateLine(dp1, l, pgp));
            FreeData(dp1);
        }
    }}}
    LeaveFunc();  return (PTRGraph*)gp;
}
PTRGraph* DrawMeshLines(PTRWindow* wp, PTRData* data,
                        int l, PGP pgp)
{
    PTRGraph* gp;
    EnterFunc("DrawMeshLines");
    if((gp=CreateMeshLines(data, l, pgp))) AddPTRGraph(wp, gp);
    LeaveFunc();  return gp;
}
PTRGraph* CreateString(double x1, double x2, double x3,
                       double v, char* str, PGP pgp)
{
    PTRString* gp;
    EnterFunc("CreateString");
    gp=(PTRString*)malloc(sizeof(PTRString));
    gp->type=STRING;
    gp->vis=TRUE;
    gp->x[AX_X1]=x1;  gp->x[AX_X2]=x2;
    gp->x[AX_X3]=x3;  gp->x[AX_V] =v;
    gp->str=(char*)malloc((strlen(str)+1)*sizeof(char));
    strcpy(gp->str, str);
    gp->pgp=pgp;
    LeaveFunc();  return (PTRGraph*)gp;
}
PTRGraph* DrawString(PTRWindow* wp, double x1, double x2,
                     double x3, double v, char* str, PGP pgp)
{
    PTRGraph* gp;
    EnterFunc("DrawString");
    if((gp=CreateString(x1, x2, x3, v, str, pgp)))
            AddPTRGraph(wp, gp);
    LeaveFunc();  return gp;
}

PTRGraph* CreateFixString(double x, double y, char* str, PGP pgp)
{
    PTRFixString* gp;
    EnterFunc("CreateFixString");
    gp=(PTRFixString*)malloc(sizeof(PTRFixString));
    gp->type=FIXSTRING;
    gp->vis=TRUE;
    gp->x=x;  gp->y=y;
    gp->str=(char*)malloc((strlen(str)+1)*sizeof(char));
    strcpy(gp->str, str);
    gp->pgp=pgp;
    LeaveFunc();  return (PTRGraph*)gp;
}
PTRGraph* DrawFixString(PTRWindow* wp, double x, double y,
                        char* str, PGP pgp)
{
    PTRGraph* gp;
    EnterFunc("DrawFixString");
    if((gp=CreateFixString(x, y, str, pgp))) AddPTRGraph(wp, gp);
    LeaveFunc();  return gp;
}
PTRGraph* CreateContourLine(PTRData* field, int l,
                            char dir, int nn, double val, PGP pgp)
{
    int i, n;
    PTRData** dts;
    PTRGraphSet* gp;
    EnterFunc("CreateContourLine");
    dts=ToContourData(field, l, dir, nn, val, &n);
    if(n==0) return NULL;
    gp=CreateGraphSet();
    for(i=0; i<n; i++){
        AddToGraphSet(gp, CreateLine(dts[i], 1, pgp));
        FreeData(dts[i]);
    }
    free(dts);
    LeaveFunc();  return (PTRGraph*)gp;
}
PTRGraph* DrawContourLine(PTRWindow* wp, PTRData* field, int l,
                          char dir, int nn, double val, PGP pgp)
{
    PTRGraph* gp;
    EnterFunc("DrawContourLine");
    if((gp=CreateContourLine(field, l, dir, nn, val, pgp)))
            AddPTRGraph(wp, gp);
    LeaveFunc();  return gp;
}
#ifdef DEBUG
PTRGraph* CreateArrows(PTRData* data, int l1, int l2, int l3,
                       double exp, PGP pgp)
{
    PTRData* dp;
    PTRArrows* gp;
    EnterFunc("CreateArrows");
    if(data){
        dp=data;
    }else if(!(dp=CurrentData())){
        SetError(ERROR, E_NoData);
        LeaveFunc();  return NULL;
    }
    if(l1<1 || l2<1){
        SetError(ERROR, E_CompNumReq);
        LeaveFunc();  return NULL;
    }else if(dp->nv<l1 || dp->nv<l2 || dp->nv<l3){
        SetError(ERROR, E_NoEnoughLeng);
        LeaveFunc();  return NULL;
    }else if(dp->dim<2){
        SetError(ERROR, E_NoEnoughDim);
        LeaveFunc();  return NULL;
    }
    gp=(PTRArrows*)malloc(sizeof(PTRArrows));
    gp->type=ARROWS;
    if(dp->type==GRID){
        int ii, i, j, k, nx, nv;
        double* x1=NULL;  double* x2=NULL;  double* x3=NULL;
        double** v;
        nx=dp->nx1*dp->nx2*dp->nx3;
        x1=(double*)malloc(nx*sizeof(double));
        x2=(double*)malloc(nx*sizeof(double));
        if(2<dp->dim){
            x3=(double*)malloc(nx*sizeof(double));
            nv=3;
        }else{
            nv=2;
        }
        v=(double**)malloc(nv*sizeof(double*));
        for(i=0; i<nv; i++)
                v[i]=(double*)malloc(nx*sizeof(double));
        ii=0;
        for(k=0; k<dp->nx3; k++){
        for(j=0; j<dp->nx2; j++){
        for(i=0; i<dp->nx1; i++){
            x1[ii]=dp->x1[i];
            x2[ii]=dp->x2[j];
            if(2<dp->dim) x3[ii]=dp->x3[k];
            v[0][ii]=ExtractV(dp, i, j, k, l1);
            v[1][ii]=ExtractV(dp, i, j, k, l2);
            if(2<dp->dim){
                if(0<l3){
                    v[2][ii]=ExtractV(dp, i, j, k, l3);
                }else{
                    v[2][ii]=0.0;
                }
            }
            ii++;
        }}}
        gp->dp=CreateSeqData(dp->dim, nx, nv, x1, x2, x3, v);
        free(x1);  free(x2);  if(x3) free(x3);
        for(i=0; i<nv; i++) free(v[i]);
        free(v);
    }else if(dp->type==SEQ){
        int i, ii, nv;  int nx=dp->nx1;
        double** v;
        if(2<dp->dim){
            nv=3;
        }else{
            nv=2;
        }
        v=(double**)malloc(nv*sizeof(double*));
        v[0]=dp->v[l1];  v[1]=dp->v[l2];
        if(2<dp->dim){
            if(0<l3){
                v[2]=dp->v[l3];
            }else{
                v[2]=(double*)malloc(nx*sizeof(double));
                for(ii=0; ii<nx; ii++) v[2][ii]=0.0;
            }
        }
        gp->dp=CreateSeqData(dp->dim, dp->nx1, nv,
                             dp->x1, dp->x2, dp->x3, v);
        if(nv==3 && l3<1) free(v[2]);
        free(v);
    }
    gp->exp=exp;  gp->pgp=pgp;
    LeaveFunc();  return (PTRGraph*)gp;
}
PTRGraph* DrawArrows(PTRWindow* wp, PTRData* data,
                     int l1, int l2, int l3, double exp, PGP pgp)
{
    PTRGraph* gp;
    EnterFunc("DrawArrows");
    if((gp=CreateArrows(data, l1, l2, l3, exp, pgp)))
            AddPTRGraph(wp, gp);
    LeaveFunc();  return gp;
}
#endif
void RemoveAxLabel(PTRWindow* wp, int ax)
{
    if(wp->label[ax]==NULL) return;
    RemoveGraph(wp, (PTRGraph*)(wp->label[ax]));
    wp->label[ax]=NULL;
}
PTRGraph* SetAxLabel(PTRWindow* wp, int ax, char* str, PGP pgp)
{
    PTRLabel* gp;
    EnterFunc("SetAxLabel");
    if(wp==NULL){
        SetError(ERROR, E_NoSuchWindow);
        LeaveFunc();  return NULL;
    }
    RemoveAxLabel(wp, ax);
    gp=(PTRLabel*)malloc(sizeof(PTRLabel));
    gp->type=LABEL;
    gp->vis=TRUE;
    gp->str=(char*)malloc((strlen(str)+1)*sizeof(char));
    strcpy(gp->str, str);
    gp->pgp=pgp;
    gp->pgp.flush=CENTER;
    gp->pgp.pgcl=DefaultForeground();
    gp->lb_dist=2.0*LB_DIST;
    gp->ax=ax;
    wp->label[ax]=gp;
    AddPTRGraph(wp, (PTRGraph*)gp);
    LeaveFunc();  return (PTRGraph*)gp;
}
void RemoveTopLabel(PTRWindow* wp){
    RemoveAxLabel(wp, AX_NL);
}
PTRGraph* SetTopLabel(PTRWindow* wp, char* str, PGP pgp){
    return SetAxLabel(wp, AX_NL, str, pgp);
}
void RemoveX1Label(PTRWindow* wp){
    RemoveAxLabel(wp, AX_X1);
}
PTRGraph* SetX1Label(PTRWindow* wp, char* str, PGP pgp){
    return SetAxLabel(wp, AX_X1, str, pgp);
}
void RemoveX2Label(PTRWindow* wp){
    RemoveAxLabel(wp, AX_X2);
}
PTRGraph* SetX2Label(PTRWindow* wp, char* str, PGP pgp){
    return SetAxLabel(wp, AX_X2, str, pgp);
}
void RemoveX3Label(PTRWindow* wp){
    RemoveAxLabel(wp, AX_X3);
}
PTRGraph* SetX3Label(PTRWindow* wp, char* str, PGP pgp){
    return SetAxLabel(wp, AX_X3, str, pgp);
}
void RemoveVLabel(PTRWindow* wp){
    RemoveAxLabel(wp, AX_V);
}
PTRGraph* SetVLabel(PTRWindow* wp, char* str, PGP pgp){
    return SetAxLabel(wp, AX_V, str, pgp);
}
void RemoveXLabel(PTRWindow* wp){
    RemoveAxLabel(wp, wp->ax1);
}
PTRGraph* SetXLabel(PTRWindow* wp, char* str, PGP pgp){
    return SetAxLabel(wp, wp->ax1, str, pgp);
}
void RemoveYLabel(PTRWindow* wp){
    RemoveAxLabel(wp, wp->ax2);
}
PTRGraph* SetYLabel(PTRWindow* wp, char* str, PGP pgp){
    return SetAxLabel(wp, wp->ax2, str, pgp);
}
void RemoveZLabel(PTRWindow* wp){
    RemoveAxLabel(wp, wp->ax3);
}
PTRGraph* SetZLabel(PTRWindow* wp, char* str, PGP pgp){
    return SetAxLabel(wp, wp->ax3, str, pgp);
}
