#include <stdlib.h>
#include <pgw.h>
#include <X11/Xutil.h>
#include "ptr.h"


PTRWindow root;

void PTRInitialize()
{
    root.pgw=NULL;
    root.wpl.next=NULL;
    SetDefaultGeometry(DFGMX, DFGMY);
    SetDefaultFrameSize((unsigned int)(A4LONG/1.5/10)*10,
                        (unsigned int)(A4SHORT/1.5/10)*10);
    SetDefaultWindowXY(0.0, 0.0);
    SetDefaultWindowSize(1.0, 1.0);
    SetDefaultColors(FGCOLOR, BGCOLOR);
    SetPSPaperSize(A4);
    SetPSOrientation(LAND);
}



int SetUp(char* serv)
{
    EnterFunc("SetUp");
    if(!PGWSetUp(serv)){
        char msg[256];
        sprintf(msg, "%s: Can't open display.\n", serv);
        SetError(FATAL, msg);
        LeaveFunc();  return FALSE;
    }
    PTRInitialize();
    LeaveFunc();  return TRUE;
}

int PSSetUp()
{
    int ret;
    ret=PGWPSSetUp();
    PTRInitialize();
    return ret;
}

void MoveWindow(PTRWindow* wp, double x, double y){
    MovePGWindow(wp->pgw, x, y);
}
void ResizeWindow(PTRWindow* wp, double width, double height)
{
    EnterFunc("ResizeWindow");
    if(width<=0.0 || height<=0.0){
        SetError(ERROR, E_BadValue);
        LeaveFunc();  return;
    }
    ResizePGWindow(wp->pgw, width, height);
    LeaveFunc();
}
void MoveResizeWindow(PTRWindow* wp,
                      double x, double y,
                      double width, double height)
{
    EnterFunc("MoveResizeWindow");
    if(width<=0.0 || height<=0.0){
        SetError(ERROR, E_BadValue);
        LeaveFunc();  return;
    }
    MoveResizePGWindow(wp->pgw, x, y, width, height);
    LeaveFunc();
}
void WindowRegion(PTRWindow* wp, int rows, int cols, int num)
{
    double x, y, width, height;
    int row, col;
    EnterFunc("WindowRegion");
    if(rows<1 || cols<1 || num<1 || rows*cols<num){
        SetError(ERROR, E_BadValue);
        LeaveFunc();  return;
    }
    width =1.0/(double)cols;
    height=1.0/(double)rows;
    col=(num-1)%cols;  row=(num-1)/cols;
    x=col*width;  y=row*height;
    MoveResizePGWindow(wp->pgw, x, y, width, height);
    LeaveFunc();
}
void ResizeFrame(PTRWindow* wp,
                 unsigned int wdx, unsigned int wdy){
    ResizePGFrame(ParentPGFrame(wp->pgw), wdx, wdy);
}

void InitializeWindow(PTRWindow* wp)
{
    int i;
    wp->gpl.next=NULL;
    wp->coord=CAR;
    wp->ax1=AX_X1;  wp->ax2=AX_V;  wp->ax3=AX_NL;
    wp->pgw->dim=DIM2;  wp->pgw->coord=CAR;
    for(i=0; i<5; i++){
        wp->label[i]=NULL; wp->scale[i]=NULL;
        wp->min[i]=0.0;  wp->max[i]=1.0;  wp->size[i]=1.0;
    }
    wp->frame=NULL;
    wp->exact=FALSE;
}

PTRWindow* CreateFrame(int gmx, int gmy, unsigned int wdx,
                       unsigned int wdy, char* title)
{
    PTRWindow* wp;
    wp=(PTRWindow*)malloc(sizeof(PTRWindow));
    wp->pgw=CreatePGFrame(gmx, gmy, wdx, wdy, title);
    wp->parent=&root;
    InitializeWindow(wp);
    return wp;
}
PTRWindow* CreateSimpleFrame()
{
    char title[64];
    PTRWindow* fp;
    sprintf(title, "%s-%s", Pointer(), ShortVersion());
    fp=CreateFrame(DefaultGMX(), DefaultGMY(),
                   DefaultWDX(), DefaultWDY(), title);
    SetDefaultGeometry(DefaultGMX()+50, DefaultGMY()+50);
    return fp;
}

PTRWindow* CreateWindow(PTRWindow* parent,
                        double x, double y,
                        double wid, double hgt)
{
    PTRWindow* wp;
    wp=(PTRWindow*)malloc(sizeof(PTRWindow));
    wp->pgw=CreatePGWindow(parent->pgw, x, y, wid, hgt);
    wp->parent=parent;
    InitializeWindow(wp);
    return wp;
}

PTRWindow* CreateSimpleWindow(PTRWindow* parent){
    return CreateWindow(parent,
                        DefaultWindowX(), DefaultWindowY(),
                        DefaultWindowW(), DefaultWindowH());
}



/* show graphs */

void ShowWindow(PTRWindow* wp){
    ShowPGWindow(wp->pgw);
}
void ShowWindows(PTRWindow* wp){
    ShowPGWindows(wp->pgw);
}
void ShowAllWindows(){
    ShowAllPGWindows();
}

/* print graphs */

void PSLandscape(char* fname)
{
    EnterFunc("PSLandscape");
    if(!OverWrite(fname)){
        LeaveFunc();  return;
    }
    if(!PGPSOpen(fname)){
        char msg[MAX_STR];
        sprintf(msg, E_FileError, fname);
        SetError(ERROR, msg);
        LeaveFunc();  return;
    }
    SetPSOrientation(LAND);
    LeaveFunc();
}
void PSPortrait(char* fname)
{
    EnterFunc("PSPortrait");
    if(!OverWrite(fname)){
        LeaveFunc();  return;
    }
    if(!PGPSOpen(fname)){
        char msg[MAX_STR];
        sprintf(msg, E_FileError, fname);
        SetError(ERROR, msg);
        LeaveFunc();  return;
    }
    SetPSOrientation(PORT);
    LeaveFunc();
}

void PSPrint(PTRWindow* wp){
    EnterFunc("PSPrint");
    if(!PGPrint(wp->pgw)) SetError(ERROR, E_OpenFile);
    LeaveFunc();
}

void ShowPage(){
    EnterFunc("ShowPage");
    if(!PGShowPage()) SetError(ERROR, E_OpenFile);
    LeaveFunc();
}


/* dump window */
int DumpPTRWindowToXpm(PTRWindow* wp, char* fname)
{
    int ret;
    char msg[256];
    EnterFunc("DumpPTRWindowToXpm");
    ret=DumpPGWindowToXpm(wp->pgw, fname);
    if(ret==Unsuported){
        sprintf(msg, E_Unsuported, "Xpm");
        SetError(ERROR, msg);
        LeaveFunc();  return Unsuported;
    }
    return ret;
}
int DumpPTRWindowToGif(PTRWindow* wp, char* fname)
{
    int ret;
    char msg[256];
    EnterFunc("DumpPTRWindowToGif");
    ret=DumpPGWindowToGif(wp->pgw, fname);
    if(ret==Unsuported){
        sprintf(msg, E_Unsuported, "gif");
        SetError(ERROR, msg);
        LeaveFunc();  return Unsuported;
    }
    return ret;
}


/* event handling */

void WaitKeyPress(char* key, PTRWindow* wp, PTRWindow* parent)
{
    KeySym keysym;
    XEvent e;
    if(key==NULL) key=(char*)malloc(sizeof(char));
    *key='\0';
    while(PGWWaitKeyPress(&e)){
        switch(e.type){
          case KeyPress:
            /* if(wp) wp = wtowp(e.xkey.window);*/
            XLookupString((XKeyEvent*)(&e), key, 1, &keysym, NULL);
            return;
        }
    }
}

void PendingKeyPress(char* key, PTRWindow* wp, PTRWindow* parent)
{
    KeySym keysym;
    XEvent e;
    if(key==NULL) key=(char*)malloc(sizeof(char));
    *key='\0';
    while(PGWPendingKeyPress(&e)){
        switch(e.type){
          case KeyPress:
            /* if(wp) wp = wtowp(e.xkey.window);*/
            XLookupString((XKeyEvent*)(&e), key, 1, &keysym, NULL);
            return;
        }
    }
}
