#include <stdlib.h>
#include <memory.h>
#include "pgw.h"


PGFontList ftlist = {0.0, 0, 0, 0.0, NULL};
extern PGWindow pgroot;


void PushPGFontList(double pt, int fmly, int sl, double sub)
{
    PGFontList* fl;
    fl=(PGFontList*)malloc(sizeof(PGFontList));
    fl->pt=pt;  fl->fmly=fmly;  fl->sl=sl; fl->sub=sub;
    fl->next=ftlist.next;
    ftlist.next=fl;
}
int PopPGFontList()
{
    PGFontList* fl;
    fl=ftlist.next;
    if(fl == NULL || fl->next == NULL) return FALSE;
    ftlist.next=fl->next;
    free(fl);
    return TRUE;
}
PGFontList CurrentPGFont()
{
    PGFontList ret={0.0, 0, 0, 0.0, NULL};
    if(ftlist.next) return *(ftlist.next);
    return ret;
}
void DuplicateFontList(){
    PGFontList fl;
    fl=CurrentPGFont();
    PushPGFontList(fl.pt, fl.fmly, fl.sl, fl.sub);
}
double CurrentFontPoint(){
    if(ftlist.next) return ftlist.next->pt;
    return 0.0;
}
void SetFontFamily(int fmly){
    if(ftlist.next) ftlist.next->fmly=fmly;
}
void SetSlant(int sl){
    if(ftlist.next) ftlist.next->sl=sl;
}
void MultipleFontPoint(double mul){
    if(ftlist.next) ftlist.next->pt*=mul;
}
void AddSubscript(double a){
    if(ftlist.next) ftlist.next->sub+=a;
}
void ClearFontList(){
    PGFontList* fl;
    PGFontList* fl1;
    fl=ftlist.next;
    while(fl){
        fl1=fl->next;  free(fl);  fl=fl1;
    }
    ftlist.next=NULL;
}


/* make PGObject */

PGPoints* CreatePGPoints(int num, double* x1, double* x2,
                         double* x3, PGP pgp)
{
    PGPoints* pg;
    pg=(PGPoints*)malloc(sizeof(PGObject));
    pg->type=PGPOINTS;
    pg->num=num;
#ifdef USE_MEMORY_COPY
    pg->x1=(double*)malloc(num*sizeof(double));
    pg->x2=(double*)malloc(num*sizeof(double));
    memcpy(pg->x1, x1, num*sizeof(double));
    memcpy(pg->x2, x2, num*sizeof(double));
#else
    pg->x1=x1;  pg->x2=x2;
#endif
    if(x3){
#ifdef USE_MEMORY_COPY
        pg->x3=(double*)malloc(num*sizeof(double));
        memcpy(pg->x3, x3, num*sizeof(double));
#else
        pg->x3=x3;
#endif
    }else{
        pg->x3=NULL;
    }
    pg->pgp=pgp;
    return pg;
}
PGLine* CreatePGLine(int num, double* x1, double* x2, double* x3,
                     PGP pgp)
{
    PGLine* pg;
    pg=(PGLine*)malloc(sizeof(PGObject));
    pg->type=PGLINE;
    pg->num=num;
#ifdef USE_MEMORY_COPY
    pg->x1=(double*)malloc(num*sizeof(double));
    pg->x2=(double*)malloc(num*sizeof(double));
    memcpy(pg->x1, x1, num*sizeof(double));
    memcpy(pg->x2, x2, num*sizeof(double));
#else
    pg->x1=x1;  pg->x2=x2;
#endif
    if(x3){
#ifdef USE_MEMORY_COPY
        pg->x3=(double*)malloc(num*sizeof(double));
        memcpy(pg->x3, x3, num*sizeof(double));
#else
        pg->x3=x3;
#endif
    }else{
        pg->x3=NULL;
    }
    pg->pgp=pgp;
    return pg;
}
PGCircle* CreatePGCircle(double x1, double x2, double x3, PGP pgp)
{
    PGCircle* pg;
    pg=(PGCircle*)malloc(sizeof(PGObject));
    pg->type=PGCIRCLE;
    pg->x1=x1;  pg->x2=x2;  pg->x3=x3;
    pg->pgp=pgp;
    return pg;
}
PGString* CreatePGString(double x1, double x2, double x3,
                         char* str, PGP pgp)
{
    PGString* pg;
    pg=(PGString*)malloc(sizeof(PGObject));
    pg->type=PGSTRING;
    pg->x1=x1;  pg->x2=x2;  pg->x3=x3;
    pg->str=(char*)malloc((strlen(str)+1)*sizeof(char));
    strcpy(pg->str, str);
    pg->pgp=pgp;
    return pg;
}

PGFixString* CreatePGFixString(double x, double y, char* str,
                               PGP pgp)
{
    PGFixString* pg;
    pg=(PGFixString*)malloc(sizeof(PGObject));
    pg->type=PGFIXSTRING;
    pg->x=x;  pg->y=y;
    pg->str=(char*)malloc((strlen(str)+1)*sizeof(char));
    strcpy(pg->str, str);
    pg->pgp=pgp;
    return pg;
}
PGArrows* CreatePGArrows(int num,
                         double* x1, double* x2, double* x3,
                         double* v1, double* v2, double* v3,
                         double exp, PGP pgp)
{
    int i;
    PGArrows* pg;
    pg=(PGArrows*)malloc(sizeof(PGObject));
    pg->type=PGARROWS;
    pg->num=num;
#ifdef USE_MEMORY_COPY
    pg->x1=(double*)malloc(num*sizeof(double));
    pg->x2=(double*)malloc(num*sizeof(double));
    memcpy(pg->x1, x1, num*sizeof(double));
    memcpy(pg->x2, x2, num*sizeof(double));
#else
    pg->x1=x1;  pg->x2=x2;
#endif
    pg->v1=(double*)malloc(num*sizeof(double));
    pg->v2=(double*)malloc(num*sizeof(double));
    for(i=0; i<num; i++){
        pg->v1[i]=pg->x1[i]+v1[i];
        pg->v2[i]=pg->x2[i]+v2[i];
    }
    if(x3){
#ifdef USE_MEMORY_COPY
        pg->x3=(double*)malloc(num*sizeof(double));
        memcpy(pg->x3, x3, num*sizeof(double));
#else
        pg->x3=x3;
#endif
    }else{
        pg->x3=NULL;
    }
    if(v3 && x3){
        pg->v3=(double*)malloc(num*sizeof(double));
        for(i=0; i<num; i++) pg->v3[i]=pg->x3[i]+v3[i];
    }else{
        pg->v3=NULL;
    }
    pg->exp=exp;
    pg->pgp=pgp;
    return pg;
}

/* add graph object to window */

void AddPGObject(PGWindow* pgw, PGObject* pg)
{
    PushToPGObjectList(&(pgw->pgl), pg);
#ifdef DEBUG
    pgl=(PGObjectList*)malloc(sizeof(PGObjectList));
    pgl->pg=pg;
    pgl->next=pgw->pgl.next;
    pgw->pgl.next=pgl;
#endif
}
void AddPGObjects(PGWindow* pgw, PGObjectList* pgl)
{
    PGObjectList* pgl1;
    if(pgl==NULL) return;
    pgl1=pgl->next;
    while(pgl1){
        AddPGObject(pgw, pgl1->pg);
        pgl1=pgl1->next;
    }
}

void RemovePGObject(PGWindow* pgw, PGObject* pg)
{
    PGObjectList* pgl;
    PGObjectList* pgl1;
    pgl=pgl1=&(pgw->pgl);
    while(pgl!=NULL && pgl->pg!=pg){
        pgl1=pgl;  pgl=pgl->next;
    }
    if(pgl==NULL) return;
    FreePGObject(pg);
    pgl1->next=pgl->next;
    FreePGObjectList(pgl);
}
void RemovePGObjects(PGWindow* pgw, PGObjectList* pgl)
{
    PGObjectList* pgl1;
    if(pgl==NULL) return;
    pgl1=pgl->next;
    while(pgl1){
        RemovePGObject(pgw, pgl1->pg);
        pgl1=pgl1->next;
    }
    FreePGObjectLists(pgl);
}

void PushToPGObjectList(PGObjectList* pglist, PGObject* pg)
{
    PGObjectList* pgl;
    if(pglist==NULL) return;
    pgl=(PGObjectList*)malloc(sizeof(PGObjectList));
    pgl->next=pglist->next;
    pgl->pg=pg;
    pglist->next=pgl;
}

/* set window properties */

void SetPGWindowX1Limit(PGWindow* pgw, double x1min, double x1max){
    if(x1max <= x1min) return;
    pgw->x1min=x1min;  pgw->x1max=x1max;  pgw->x1size=x1max-x1min;
}
void SetPGWindowX2Limit(PGWindow* pgw, double x2min, double x2max){
    if(x2max <= x2min) return;
    pgw->x2min=x2min;  pgw->x2max=x2max;  pgw->x2size=x2max-x2min;
}
void SetPGWindowX3Limit(PGWindow* pgw, double x3min, double x3max){
    if(x3max <= x3min) return;
    pgw->x3min=x3min;  pgw->x3max=x3max;  pgw->x3size=x3max-x3min;
}
void SetPGWindowX2_X1(PGWindow* pgw, double x2_x1){
    pgw->x2_x1=x2_x1;
}
void SetPGWindowX3_X1(PGWindow* pgw, double x3_x1){
    pgw->x3_x1=x3_x1;
}


/* set coordinate */

void SetPGWindowCoord(PGWindow* pgw, int coord)
{
    switch(coord){
      case CAR:
      case CYL:
        pgw->coord=coord;
        break;
      default:
        return;
    }
}
void SetPGWindowDimension(PGWindow* pgw, int dim)
{
    switch(dim){
      case DIM2:
      case DIM3:
        pgw->dim=dim;
        break;
      default:
        return;
    }
}

/* window properties */
int Is3D(PGWindow* pgw){
    return (pgw->dim==DIM3);
}
int IsPGFrame(PGWindow* pgw){
    return (pgw->parent==&pgroot);
}
PGWindow* ParentPGWindow(PGWindow* pgw){
    return pgw->parent;
}
PGWindow* ParentPGFrame(PGWindow* pgw){
    PGWindow* parent;
    parent=pgw;
    while(!IsPGFrame(parent))
            parent=ParentPGWindow(parent);
    return parent;
}

void SetPhiAngle(PGWindow* pgw, double ang){
    pgw->phi=ang;
    while(pgw->phi<=0.0) pgw->phi+=2*PI;
    while(2*PI<pgw->phi) pgw->phi-=2*PI;
}
void SetThetaAngle(PGWindow* pgw, double ang){
    pgw->theta=ang;
    if(pgw->theta<=Theta_Min) pgw->theta=Theta_Min;
    if(Theta_Max<=pgw->theta) pgw->theta=Theta_Max;

}void IncreasePhi(PGWindow* pgw, double ang){
    pgw->phi+=ang;
    while(pgw->phi<=0.0) pgw->phi+=2*PI;
    while(2*PI<pgw->phi) pgw->phi-=2*PI;
}
void IncreaseTheta(PGWindow* pgw, double ang){
    pgw->theta+=ang;
    if(pgw->theta<=Theta_Min) pgw->theta=Theta_Min;
    if(Theta_Max<=pgw->theta) pgw->theta=Theta_Max;
}
void SetMag(PGWindow* pgw, double mag){
    if(mag<=0.0) return;
    pgw->mag=mag;
}
void SetOrgx(PGWindow* pgw, double ox){
    pgw->orgx=ox;
}
void SetOrgy(PGWindow* pgw, double oy){
    pgw->orgy=oy;
}


/* color operation */

/* h: hue        (0-360) */
/* s: saturation (0-1)   */
/* v: value      (0-1)   */

void HSVToRGB(double h, double s, double v, int* r, int* g, int* b)
{
    double* max;  double* mid;  double* min;
    double rr, gg, bb;
    double a;

    if(0.0<=h && h<60.0){
        max=&rr;  mid=&gg;  min=&bb;  a=h;
    }else if(h<120.0){
        max=&gg;  mid=&rr;  min=&bb;  a=120.0-h;
    }else if(h<180.0){
        max=&gg;  mid=&bb;  min=&rr;  a=h-120.0;
    }else if(h<240.0){
        max=&bb;  mid=&gg;  min=&rr;  a=240.0-h;
    }else if(h<300.0){
        max=&bb;  mid=&rr;  min=&gg;  a=h-240.0;
    }else if(h<=360.0){
        max=&rr;  mid=&bb;  min=&gg;  a=360.0-h;
    }
    *max=v;
    *min=(1.0-s)*(*max);
    *mid=a/60.0*(*max-*min)+(*min);
    *r=(int)(255*rr+0.5);
    *g=(int)(255*gg+0.5);
    *b=(int)(255*bb+0.5);
}

int GrayColors(int n, PGColor pgcl[])
{
    int i, v, ret;
    int* r;  int* g;  int* b;
    r=(int*)malloc(n*sizeof(int));
    g=(int*)malloc(n*sizeof(int));
    b=(int*)malloc(n*sizeof(int));
    for(i=0; i<n; i++){
        v=(int)(i/(n-1.0)*255+0.5);
        r[i]=v;  g[i]=v;  b[i]=v;
    }
    ret=PGGetColors(n, r, g, b, pgcl);
    free(r);  free(g);  free(b);
    return ret;
}
