#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <ptr.h>
#include <math.h>
#include "pointer.h"


ShortWordList shortcom[] =
{
    {"drawstr", "drawstring"},
    {"exrat", "exactratio"},
    {"frame", "scale"},
    {"input", "open"},
    {"lim", "limit"},
    {"newwin", "newwindow"},
    {"nf", "newframe"},
    {"nw", "newwindow"},
    {"pr", "print"},
    {"psl", "psland"},
    {"psp", "psport"},
    {"q", "quit"},
    {"reg", "region"},
    {"sp", "showpage"},
    {"toplab", "toplabel"},
    {"vlab",  "vlabel"},
    {"vlim",  "vlimit"},
    {"x1lab", "x1label"},
    {"x1lim", "x1limit"},
    {"x2lab", "x2label"},
    {"x2lim", "x2limit"},
    {"x3lab", "x3label"},
    {"x3lim", "x3limit"},
    {"xlab", "xlabel"},
    {"xlim", "xlimit"},
    {"ylab", "ylabel"},
    {"ylim", "ylimit"},
    {"zlab", "zlabel"},
    {"zlim", "zlimit"}
};
int nshortcom=sizeof(shortcom)/sizeof(*shortcom);

CommandList commands[] =
{
    {"alias", C_Alias, 0},
    {"axis", C_Axis, 0},
    {"axisratio", C_AxisRatio, 0},
    {"break", C_Break, 0},
    {"baselinewidth", C_BaseLineWidth, 0},
    {"cd", C_Cd, 0},
    {"clear", C_Clear, 0},
    {"color", C_Color, 0},
    {"continue", C_Continue, 0},
    {"contour", C_Contour, 0},
    {"coord", C_Coord, 0},
    {"cos", C_Cos, 0},
    {"data", C_Data, 0},
    {"defaultbackground", C_DefaultBackground, 0},
    {"defaultfont", C_DefaultFont, 0},
    {"defaultforeground", C_DefaultForeground, 0},
    {"defaultslant", C_DefaultSlant, 0},
    {"draw", C_Draw, 0},
    {"drawstring", C_DrawString, 0},
    {"dump", C_Dump, 0},
    {"echo", C_Echo, 0},
    {"exactratio", C_ExactRatio, 0},
    {"exp", C_Exp, 0},
    {"fgets", C_FGetS, 0},
    {"file", C_File, 0},
    {"fread", C_FRead, 0},
    {"fseek", C_FSeek, 0},
    {"getcolor", C_GetColor, 0},
    {"gets", C_GetS, 0},
    {"global", C_Global, 1},
    {"graycolors", C_GrayColors, 0},
    {"isundef", C_IsUndef, 1},
    {"limit", C_Limit, 0},
    {"linewidth", C_LineWidth, 0},
    {"log", C_Log, 0},
    {"log10", C_Log10, 0},
    {"maximum", C_Maximum, 0},
    {"meshlines", C_MeshLines, 0},
    {"minimum", C_Minimum, 0},
    {"minmax", C_MinMax, 0},
    {"newframe", C_NewFrame, 0},
    {"newwindow", C_NewWindow, 0},
    {"open", C_Open, 0},
    {"plot", C_Plot, 0},
    {"plotsize", C_PlotSize, 0},
    {"print", C_Print, 0},
    {"printf", C_PrintF, 0},
    {"psland", C_PSLand, 0},
    {"psport", C_PSPort, 0},
    {"quit", C_Quit, 0},
    {"raise", C_Raise, 0},
    {"read", C_Read, 0},
    {"redraw", C_Redraw, 0},
    {"region", C_Region, 0},
    {"resizeframe", C_ResizeFrame, 0},
    {"restoredata", C_RestoreData, 0},
    {"return", C_Return, 0},
    {"savedata", C_SaveData, 0},
    {"scale", C_Scale, 0},
    {"showpage", C_ShowPage, 0},
    {"sin", C_Sin, 0},
    {"split", C_Split, 0},
    {"sprintf", C_SPrintF, 0},
    {"sqrt", C_Sqrt, 0},
    {"system", C_System, 0},
    {"tan", C_Tan, 0},
    {"toint", C_ToInt, 0},
    {"togif", C_ToGif, 0},
    {"toplabel", C_TopLabel, 0},
    {"toxpm", C_ToXpm, 0},
    {"version", C_Version, 0},
    {"vlabel",  C_VLabel, 0},
    {"vlimit",  C_VLimit, 0},
    {"waitkey", C_WaitKey, 0},
    {"x1label", C_X1Label, 0},
    {"x1limit", C_X1Limit, 0},
    {"x2label", C_X2Label, 0},
    {"x2limit", C_X2Limit, 0},
    {"x3label", C_X3Label, 0},
    {"x3limit", C_X3Limit, 0},
    {"xlabel", C_XLabel, 0},
    {"xlimit", C_XLimit, 0},
    {"ylabel", C_YLabel, 0},
    {"ylimit", C_YLimit, 0},
    {"zlabel", C_ZLabel, 0},
    {"zlimit", C_ZLimit, 0}
};
int ncommands=sizeof(commands)/sizeof(*commands);




CommandList NameToCommand(char* name)
{
    ShortWordList* sw;
    CommandList* cm;
    CommandList ret = {"", NULL};
    char nm[64];
    if(name[0]=='\\') name++;
    strcpy(nm, name);
    for(sw=shortcom; sw<shortcom+nshortcom; sw++){
        if(strcmp(name, sw->sname) == 0){
            strcpy(nm, sw->name);
            break;
        }
    }
    for(cm=commands; cm<commands+ncommands; cm++){
        if(strcmp(nm, cm->name)) continue;
        return *cm;
    }
    return ret;
}

int IsVarCommand(CommandList cm){
    return cm.var;
}

Element* C_Alias(int argc, Element* argv[])
{
    Element* el;
    List* le;
    if(argc==1){
        el=VarNameToElement(V_Alias);
        if(!el || el->type!=LIST){
            printf("No aliases.\n");
            return NULL;
        }
        le=el->list.next;
        while(le){
            printf("%-8s", ToString(le->el));
            le=le->next;
            if(le){
                printf("%s\n", ToString(le->el));
            }else{
                printf("\n");
                return NULL;
            }
            le=le->next;
        }
        return NULL;
    }else if(argc==2){
        el=GetAsArrayValue(V_Alias, argv[1]);
        if(!el) return NULL;
        printf("%-8s", ToString(argv[1]));
        printf("%s\n", ToString(el));
        return NULL;
    }
    el=ListElement(argc-2, argv+2);
    AddToAsArrayVar(V_Alias, CopyElement(argv[1]), el);
    return NULL;
}
Element* C_Axis(int argc, Element* argv[])
{
    int i, n=0;
    int ax[]={AX_X1, AX_V, AX_NL};
    PTRWindow* wp=NULL;
    for(i=1; i<argc; i++){
        if(argv[i]->type==WINDOW){
            wp=argv[i]->window.wp;
        }else if(argv[i]->type==STR){
            if(2<n) continue;
            ax[n]=NameToAxis(argv[i]->string.val);
            if(ax[n]<0 || 4<ax[n]){
                char msg[256];
                sprintf(msg, E_NoSuchAxis, ToString(argv[i]));
                CommandError(ToString(argv[0]), msg);
                return NULL;
            }
            n++;
        }
    }
    if(!wp && !(wp=ActiveWindow())){
        CommandError(ToString(argv[0]), E_NoActWin);
        return NULL;
    }
    if(n==0){
        printf("(%s, %s, %s)\n", AxisToName(wp->ax1),
               AxisToName(wp->ax2), AxisToName(wp->ax3));
        return NULL;
    }else if(n==1){
        CommandError(ToString(argv[0]), E_FewArgs);
        return NULL;
    }else{
        SetAxis(wp, ax[0], ax[1], ax[2]);
        ShowWindow(wp);
    }
    return NULL;
}
Element* C_AxisRatio(int argc, Element* argv[])
{
    int i, n=0;
    double a[3];
    PTRWindow* wp=NULL;
    for(i=1; i<argc; i++){
        if(argv[i]->type==DOUBLE && n<3){
            a[n]=argv[i]->real.val;  n++;
        }else if(argv[i]->type==INT && n<3){
            a[n]=argv[i]->integer.val;  n++;
        }else if(argv[i]->type==WINDOW){
            wp=argv[i]->window.wp;
        }else{
            CommandError(ToString(argv[0]), E_BadArgs);
            return NULL;
        }
    }
    if(n<2){
        CommandError(ToString(argv[0]), E_FewArgs);
        return NULL;
    }
    if(!wp && !(wp=ActiveWindow())){
        CommandError(ToString(argv[0]), E_NoActWin);
        return NULL;
    }
    if(a[0]<=0.0){
        char msg[256];
        sprintf(msg, E_BadArgD, a[0]);
        CommandError(ToString(argv[0]), msg);
        return NULL;
    }
    SetY_X(wp, a[1]/a[0]);
    if(n==3) SetZ_X(wp, a[2]/a[0]);
    ShowWindow(wp);
    return NULL;
}

Element* C_BaseLineWidth(int argc, Element* argv[])
{
    int i, n=0;
    double width=0.0;
    if(argc==1){
        printf("%s: %lf\n", ToString(argv[0]), BaseLineWidth());
        return NULL;
    }
    for(i=1; i<argc; i++){
        if(argv[i]->type==DOUBLE){
            width=argv[i]->real.val;
        }else if(argv[i]->type==INT){
            width=(double)(argv[i]->integer.val);
        }else{
            CommandError(ToString(argv[0]), E_BadArgs);
            return NULL;
        }
    }
    SetBaseLineWidth(width);
    ShowAllWindows();
    return NULL;
}
Element* C_Break(int argc, Element* argv[])
{
    SetBreakSignal();
    return NULL;
}
Element* C_Cd(int argc, Element* argv[])
{
    char dir1[256];
    char* dir;
    if(argc==1){
        strcpy(dir1, "~");
    }else{
        strcpy(dir1, ToString(argv[1]));
    }
    dir=ExpandHomeDirectory(dir1);
    if(chdir(dir)!=0) perror(dir);
    if(dir!=dir1) free(dir);
    return NULL;
}
Element* C_Clear(int argc, Element* argv[])
{
    int i;
    PTRWindow* wp=NULL;
    for(i=argc-1; 0<i; i--){
        if(argv[i]->type==WINDOW) wp=argv[i]->window.wp;
    }
    if(!wp && !(wp=ActiveWindow())){
        CommandError(ToString(argv[0]), E_NoActWin);
        return NULL;
    }
    ClearWindow(wp);
    ShowWindow(wp);
    return NULL;
}
Element* C_Color(int argc, Element* argv[])
{
    int i, n=0;
    int a[3];
    Element* el=NULL;
    if(argc==1){
        CommandError(ToString(argv[0]), E_NoColName);
        return NULL;
    }
    for(i=1; i<argc; i++){
        if(argv[i]->type==STR){
            el=argv[i];
        }else if(argv[i]->type==LIST){
            el=argv[i];
        }else if(argv[i]->type==INT && n<3){
            a[n]=argv[i]->integer.val;  n++;
        }
    }
    if(!el){
        if(3<=n){
            SetColor(a[0], a[1], a[2]);
        }else{
            CommandError(ToString(argv[0]), E_BadArgs);
            return NULL;
        }
    }else if(el->type==LIST){
        a[0]=ToInteger(NthElement(1, (List*)argv[1]));
        a[1]=ToInteger(NthElement(2, (List*)argv[1]));
        a[2]=ToInteger(NthElement(3, (List*)argv[1]));
        SetColor(a[0], a[1], a[2]);
    }else if(el->type==STR && el->string.val[0]=='#'){
        char buf[3]="ff";
        for(i=0; i<3; i++){
            buf[0]=argv[1]->string.val[2*i+1];
            buf[1]=argv[1]->string.val[2*i+2];
            if(!sscanf(buf, "%x", a+i)){
                char msg[256];
                sprintf(msg, E_BadArg, el->string.val);
                CommandError(ToString(argv[0]), msg);
                return NULL;
            }
        }
        SetColor(a[0], a[1], a[2]);
    }else if(el->type==STR){
        SetNamedColor(el->string.val);
    }else{
        CommandError(ToString(argv[0]), E_BadArgs);
        return NULL;
    }
    return NULL;
}
Element* C_Continue(int argc, Element* argv[])
{
    SetContinueSignal();
    return NULL;
}
Element* C_Contour(int argc, Element* argv[])
{
    int i, l=1, nn=0, ltype;
    char dir='k';
    char* str=NULL;
    PGP pgp=GetPGP();
    PTRData* dp=NULL;
    PTRWindow* wp=NULL;
    Element* vals=NULL;
    for(i=1; i<argc; i++){
        if(argv[i]->type==WINDOW){
            wp=argv[i]->window.wp;
        }else if(argv[i]->type==DATA){
            UpdateDataElement((Data*)(argv[i]));
            dp=argv[i]->data.dp;
        }else if(argv[i]->type==STR){
            if(argv[i]->string.val[1]=='\0'){
                dir=argv[i]->string.val[0];
            }else if(argv[i]->string.val[0]=='-'
                     && argv[i]->string.val[2]=='\0'){
                dir=argv[i]->string.val[1];
            }else{
                str=argv[i]->string.val;
                continue;
            }
            i++;
            if(argc<=i){
                CommandError(ToString(argv[0]), E_FewArgs);
                return NULL;
            }else if(argv[i]->type!=INT){
                char msg[256];
                sprintf(msg, E_BadArg, ToString(argv[i]));
                CommandError(ToString(argv[0]), msg);
                return NULL;
            }
            nn=argv[i]->integer.val;
        }else if(argv[i]->type==INT){
            l=argv[i]->integer.val;
        }else if(argv[i]->type==DOUBLE
                 || argv[i]->type==DARRAY
                 || argv[i]->type==LIST){
            vals=argv[i];
        }else if(argv[i]->type==COLOR){
            pgp.pgcl=argv[i]->color.pgcl;
        }
    }
    if(str){
        if(!LTNToLT(str, &ltype)){
            char msg[256];
            sprintf(msg, E_NoLineType, str);
            CommandError(ToString(argv[0]), msg);
            return NULL;
        }
    }else{
        ltype=DefaultLineType();
    }
    if(!wp && !(wp=ActiveWindow())){
        CommandError(ToString(argv[0]), E_NoActWin);
        return NULL;
    }
    if(!dp && !(dp=DefaultData())){
        CommandError(ToString(argv[0]), E_NoDataSp);
        return NULL;
    }
    pgp.ltype=ltype;
    if(vals==NULL){
        int nv=10;
        double v, min, max, size;
        if(!ExtractVRange(dp, l, &min, &max)) return NULL;
        size=max-min;
        for(i=0; i<nv; i++){
            v=min+(i+0.5)*size/(nv);
            DrawContourLine(wp, dp, l, dir, nn, v, pgp);
        }
    }else if(vals->type==DOUBLE){
        DrawContourLine(wp, dp, l, dir, nn, vals->real.val, pgp);
    }else if(vals->type==DARRAY){
        for(i=0; i<vals->darray.num; i++){
            DrawContourLine(wp, dp, l, dir, nn,
                            vals->darray.val[i], pgp);
        }
    }else if(vals->type==LIST){
        int num;
        double v;
        Element* el;
        num=ListLength((List*)vals);
        for(i=0; i<num; i++){
            el=NthElement(i+1, (List*)vals);
            if(el==NULL){
                continue;
            }else if(el->type==DOUBLE){
                v=vals->real.val;
            }else if(el->type==INT){
                v=vals->integer.val;
            }else{
                continue;
            }
            DrawContourLine(wp, dp, l, dir, nn, v, pgp);
        }
    }
    ShowWindow(wp);
    return NULL;
}
Element* C_Coord(int argc, Element* argv[])
{
    int i;
    int coord=0;
    PTRWindow* wp=NULL;
    for(i=1; i<argc; i++){
        if(argv[i]->type==WINDOW){
            wp=argv[i]->window.wp;
        }else if(argv[i]->type==STR){
            coord=NameToCoord(argv[i]->string.val);
            if(!coord){
                char msg[256];
                sprintf(msg, E_NoSuchCoord, ToString(argv[i]));
                CommandError(ToString(argv[0]), msg);
                return NULL;
            }
        }
    }
    if(!wp && !(wp=ActiveWindow())){
        CommandError(ToString(argv[0]), E_NoActWin);
        return NULL;
    }
    if(coord==0){
        printf("%s\n", CoordToName(wp->coord));
        return NULL;
    }else{
        SetCoord(wp, coord);
        ShowWindow(wp);
    }
    return NULL;
}
Element* C_Cos(int argc, Element* argv[])
{
    int i;
    extern double cos();
    List* list;
    list=ToList(NULL);
    if(argc==1) return NULL;
    if(argc==2) return MathElement(argv[1], cos);
    for(i=argc-1; 0<i; i--){
        PushToList(list, MathElement(argv[i], cos));
    }
    return (Element*)list;
}

Element* C_Data(int argc, Element* argv[])
{
    Element* ret;
    ret=DataElement(NULL);
    DefaultVariable(V_Data);
    return ret;
}

Element* C_DefaultFont(int argc, Element* argv[])
{
    int i;
    char* str=NULL;
    for(i=argc-1; 0<i; i--){
        if(argv[i]->type==STR){
            str=argv[i]->string.val;
        }else{
            char msg[256];
            sprintf(msg, E_BadArg, ToString(argv[i]));
            CommandError(ToString(argv[0]), msg);
            return NULL;
        }
    }
    if(str){
        int font;
        if(FontNameToFont(str, &font)){
            SetDefaultFontFamily(font);
        }else{
            char msg[256];
            sprintf(msg, E_NoFont, str);
            CommandError(ToString(argv[0]), msg);
            return NULL;
        }
    }else{
        printf("%s: %s\n", ToString(argv[0]),
               FontToFontName(DefaultFontFamily()));
    }
    return NULL;
}
Element* C_DefaultSlant(int argc, Element* argv[])
{
    int i;
    char* str=NULL;
    for(i=argc-1; 0<i; i--){
        if(argv[i]->type==STR){
            str=argv[i]->string.val;
        }else{
            char msg[256];
            sprintf(msg, E_BadArg, ToString(argv[i]));
            CommandError(ToString(argv[0]), msg);
            return NULL;
        }
    }
    if(str){
        int slant;
        if(SlantNameToSlant(str, &slant)){
            SetDefaultSlant(slant);
        }else{
            char msg[256];
            sprintf(msg, E_NoSlant, str);
            CommandError(ToString(argv[0]), msg);
            return NULL;
        }
    }else{
        printf("%s: %s\n", ToString(argv[0]),
               SlantToSlantName(DefaultSlant()));
    }
    return NULL;
}
Element* C_DefaultForeground(int argc, Element* argv[])
{
    int i, n=0;
    int a[3];
    PGColor pgcl;
    Element* el=NULL;
    if(argc==1){
        CommandError(ToString(argv[0]), E_NoColName);
        return NULL;
    }
    for(i=1; i<argc; i++){
        if(argv[i]->type==STR){
            el=argv[i];
        }else if(argv[i]->type==LIST){
            el=argv[i];
        }else if(argv[i]->type==INT && n<3){
            a[n]=argv[i]->integer.val;  n++;
        }
    }
    if(!el){
        if(3<=n){
            if(!GetColor(a[0], a[1], a[2], &pgcl)){
                CommandError(ToString(argv[0]), E_AllocCol);
                return NULL;
            }
            SetDefaultForeground(pgcl);
        }else{
            CommandError(ToString(argv[0]), E_BadArgs);
            return NULL;
        }
    }else if(el->type==LIST){
        a[0]=ToInteger(NthElement(1, (List*)argv[1]));
        a[1]=ToInteger(NthElement(2, (List*)argv[1]));
        a[2]=ToInteger(NthElement(3, (List*)argv[1]));
        if(!GetColor(a[0], a[1], a[2], &pgcl)){
            CommandError(ToString(argv[0]), E_AllocCol);
            return NULL;
        }
        SetDefaultForeground(pgcl);
    }else if(el->type==STR && el->string.val[0]=='#'){
        char buf[3]="ff";
        for(i=0; i<3; i++){
            buf[0]=argv[1]->string.val[2*i+1];
            buf[1]=argv[1]->string.val[2*i+2];
            if(!sscanf(buf, "%x", a+i)){
                char msg[256];
                sprintf(msg, E_BadArg, el->string.val);
                CommandError(ToString(argv[0]), msg);
                return NULL;
            }
        }
        if(!GetColor(a[0], a[1], a[2], &pgcl)){
            CommandError(ToString(argv[0]), E_AllocCol);
            return NULL;
        }
        SetDefaultForeground(pgcl);
    }else if(el->type==STR){
        if(!GetNamedColor(el->string.val, &pgcl)){
            CommandError(ToString(argv[0]), E_AllocCol);
            return NULL;
        }
        SetDefaultForeground(pgcl);
    }else{
        CommandError(ToString(argv[0]), E_BadArgs);
        return NULL;
    }
    return NULL;
}
Element* C_DefaultBackground(int argc, Element* argv[])
{
    int i, n=0;
    int a[3];
    PGColor pgcl;
    Element* el=NULL;
    if(argc==1){
        CommandError(ToString(argv[0]), E_NoColName);
        return NULL;
    }
    for(i=1; i<argc; i++){
        if(argv[i]->type==STR){
            el=argv[i];
        }else if(argv[i]->type==LIST){
            el=argv[i];
        }else if(argv[i]->type==INT && n<3){
            a[n]=argv[i]->integer.val;  n++;
        }
    }
    if(!el){
        if(3<=n){
            if(!GetColor(a[0], a[1], a[2], &pgcl)){
                CommandError(ToString(argv[0]), E_AllocCol);
                return NULL;
            }
            SetDefaultBackground(pgcl);
        }else{
            CommandError(ToString(argv[0]), E_BadArgs);
            return NULL;
        }
    }else if(el->type==LIST){
        a[0]=ToInteger(NthElement(1, (List*)argv[1]));
        a[1]=ToInteger(NthElement(2, (List*)argv[1]));
        a[2]=ToInteger(NthElement(3, (List*)argv[1]));
        if(!GetColor(a[0], a[1], a[2], &pgcl)){
            CommandError(ToString(argv[0]), E_AllocCol);
            return NULL;
        }
        SetDefaultBackground(pgcl);
    }else if(el->type==STR && el->string.val[0]=='#'){
        char buf[3]="ff";
        for(i=0; i<3; i++){
            buf[0]=argv[1]->string.val[2*i+1];
            buf[1]=argv[1]->string.val[2*i+2];
            if(!sscanf(buf, "%x", a+i)){
                char msg[256];
                sprintf(msg, E_BadArg, el->string.val);
                CommandError(ToString(argv[0]), msg);
                return NULL;
            }
        }
        if(!GetColor(a[0], a[1], a[2], &pgcl)){
            CommandError(ToString(argv[0]), E_AllocCol);
            return NULL;
        }
        SetDefaultBackground(pgcl);
    }else if(el->type==STR){
        if(!GetNamedColor(el->string.val, &pgcl)){
            CommandError(ToString(argv[0]), E_AllocCol);
            return NULL;
        }
        SetDefaultBackground(pgcl);
    }else{
        CommandError(ToString(argv[0]), E_BadArgs);
        return NULL;
    }
    return NULL;
}


Element* C_Draw(int argc, Element* argv[])
{
    int i, n=1, ltype;
    char* str=NULL;
    PGP pgp=GetPGP();
    PTRData* dp=NULL;
    PTRWindow* wp=NULL;
    for(i=argc-1; 0<i; i--){
        if(argv[i]->type==WINDOW){
            wp=argv[i]->window.wp;
        }else if(argv[i]->type==DATA){
            UpdateDataElement((Data*)(argv[i]));
            dp=argv[i]->data.dp;
        }else if(argv[i]->type==STR){
            str=argv[i]->string.val;
        }else if(argv[i]->type==INT){
            n=argv[i]->integer.val;
        }else if(argv[i]->type==COLOR){
            pgp.pgcl=argv[i]->color.pgcl;
        }
    }
    if(str){
        if(!LTNToLT(str, &ltype)){
            char msg[256];
            sprintf(msg, E_NoLineType, str);
            CommandError(ToString(argv[0]), msg);
            return NULL;
        }
    }else{
        ltype=DefaultLineType();
    }
    if(!wp && !(wp=ActiveWindow())){
        CommandError(ToString(argv[0]), E_NoActWin);
        return NULL;
    }
    if(!dp && !(dp=DefaultData())){
        CommandError(ToString(argv[0]), E_NoDataSp);
        return NULL;
    }
    pgp.ltype=ltype;
    DrawLine(wp, dp, n, pgp);
    ShowWindow(wp);
    return NULL;
}
Element* C_DrawString(int argc, Element* argv[])
{
    int i, n=0, fix=FALSE;
    char* str=NULL;
    double x[4];
    PGP pgp=GetPGP();
    List* xl=NULL;
    PTRWindow* wp=NULL;
    for(i=1; i<argc; i++){
        if(argv[i]->type==WINDOW){
            wp=argv[i]->window.wp;
        }else if(argv[i]->type==STR){
            if(strcmp(argv[i]->string.val, "-f")==0){
                fix=TRUE;
            }else if(strcmp(argv[i]->string.val, "-c")==0){
                pgp.flush=CENTER;
            }else if(strcmp(argv[i]->string.val, "-l")==0){
                pgp.flush=LEFT;
            }else if(strcmp(argv[i]->string.val, "-r")==0){
                pgp.flush=RIGHT;
            }else{
                str=argv[i]->string.val;
            }
        }else if(argv[i]->type==INT || argv[i]->type==DOUBLE){
            if(n<4){
                x[n]=ToDouble(argv[i]);
                n++;
            }
        }else if(argv[i]->type==LIST){
            xl=(List*)argv[i];
        }else if(argv[i]->type==COLOR){
            pgp.pgcl=argv[i]->color.pgcl;
        }
    }
    if(!wp && !(wp=ActiveWindow())){
        CommandError(ToString(argv[0]), E_NoActWin);
        return NULL;
    }
    if(!str && !(str=DefaultString())){
        CommandError(ToString(argv[0]), E_NoStrSp);
        return NULL;
    }
    if(xl){
        Element* tmp;
        n=ListLength(xl);
        if(4<n) n=4;
        for(i=0; i<4; i++){
            tmp=NthElement(i+1, xl);
            if(!tmp || (tmp->type!=INT && tmp->type!=DOUBLE)){
                char msg[256];
                sprintf(msg, E_BadArg, ToString(tmp));
                CommandError(ToString(argv[0]), msg);
                return NULL;
            }
            x[i]=ToDouble(tmp);
        }
    }
    if(fix){
        if(n<2){
            CommandError(ToString(argv[0]), E_FewArgs);
            return NULL;
        }
        DrawFixString(wp, x[0], x[1], str, pgp);
    }else{
        double xx[4]={0.0, 0.0, 0.0, 0.0};
        if(n==4){
            for(i=0; i<4; i++) xx[i]=x[i];
        }else if(wp->ax3){
            if(n<3){
                CommandError(ToString(argv[0]), E_FewArgs);
                return NULL;
            }
            xx[wp->ax1-1]=x[0];
            xx[wp->ax2-1]=x[1];
            xx[wp->ax3-1]=x[2];
        }else{
            if(n<2){
                CommandError(ToString(argv[0]), E_FewArgs);
                return NULL;
            }
            xx[wp->ax1-1]=x[0];
            xx[wp->ax2-1]=x[1];
        }
        DrawString(wp, xx[0], xx[1], xx[2], xx[3], str, pgp);
    }
    ShowWindow(wp);
    return NULL;
}

Element* C_Dump(int argc, Element* argv[])
{
    int i;
    char* str=NULL;
    PTRWindow* wp=NULL;
    for(i=argc-1; 0<i; i--){
        if(argv[i]->type==WINDOW){
            wp=argv[i]->window.wp;
        }else if(argv[i]->type==STR){
            str=argv[i]->string.val;
        }
    }
    if(str==NULL){
        CommandError(ToString(argv[0]), E_NoFileSp);
        return NULL;
    }
    if(!wp && !(wp=ActiveWindow())){
        CommandError(ToString(argv[0]), E_NoActWin);
        return NULL;
    }
    switch(FileSuffix(str)){
      case F_GIF:
        DumpPTRWindowToGif(wp, str);
        break;
      case F_XPM:
        DumpPTRWindowToXpm(wp, str);
        break;
      default:
        CommandError(ToString(argv[0]), E_UnknownSuf);
        return NULL;
    }
    return NULL;
}

Element* C_Echo(int argc, Element* argv[])
{
    int i;
    char* s;
    for(i=1; i<argc; i++){
        if(1<i) putchar(' ');
        s=ToString(argv[i]);
        printf("%s", s);
        if(argv[i] && argv[i]->type!=STR) free(s);
    }
    putchar('\n');
    return NULL;
}

Element* C_ExactRatio(int argc, Element* argv[])
{
    int i, set=0;
    PTRWindow* wp=NULL;
    Element* el=NULL;
    for(i=argc-1; 0<i; i--){
        if(argv[i]->type==WINDOW){
            wp=argv[i]->window.wp;
        }else{
            el=argv[i];
        }
    }
    if(!wp && !(wp=ActiveWindow())){
        CommandError(ToString(argv[0]), E_NoActWin);
        return NULL;
    }
    if(el){
        set=ToBoolean(el)-IsExactRatio(wp);
    }else{
        set=1-2*IsExactRatio(wp);
    }
    if(set==1){
        ExactRatio(wp, TRUE);
    }else if(set==-1){
        ExactRatio(wp, FALSE);
    }else{
        return NULL;
    }
    ShowWindow(wp);
    return NULL;
}

Element* C_Exp(int argc, Element* argv[])
{
    int i;
    extern double exp();
    List* list;
    list=ToList(NULL);
    if(argc==1) return NULL;
    if(argc==2) return MathElement(argv[1], exp);
    for(i=argc-1; 0<i; i--){
        PushToList(list, MathElement(argv[i], exp));
    }
    return (Element*)list;
}

Element* C_FGetS(int argc, Element* argv[])
{
    int i;
    char* buf;
    FILE* fp=NULL;
    Element* ret;
    for(i=argc-1; 0<i; i--){
        if(argv[i]->type==FP) fp=argv[i]->file.fp;
    }
    if(!fp && !(fp=DefaultFile())){
        CommandError(ToString(argv[0]), E_NoFileSp);
        return NULL;
    }
    buf=(char*)malloc(256*sizeof(char));
    ReadLine(buf, 256, fp);
    DefaultVariable(V_String);
    ret=StringElement(buf);
    free(buf);
    return ret;
}

Element* C_File(int argc, Element* argv[])
{
    int i;
    Element* el=NULL;
    char* str=NULL;
    for(i=argc-1; 0<i; i--){
        if(argv[i]->type==FP){
            el=argv[i];
        }else if(argv[i]->type==STR){
            str=argv[i]->string.val;
        }
    }
    if(str){
        char* fname;
        FILE* fp;
        fname=ExpandHomeDirectory(str);
        if(!(fp=fopen(fname, "r"))){
            char msg[256];
            sprintf(msg, E_FileError, str);
            CommandError(ToString(argv[0]), msg);
            if(fname!=str) free(fname);
            return NULL;
        }
        ProccessFile(fname, fp);
        if(fname!=str) free(fname);
        fclose(fp);
    }else{
        if(!el && !(el=DefaultFileElement())){
            CommandError(ToString(argv[0]), E_NoFileSp);
            return NULL;
        }
        ProccessFile(el->file.name, el->file.fp);
    }
    return NULL;
}

Element* C_FRead(int argc, Element* argv[])
{
    int i, n=1, size;
    void* buf;
    char* type=NULL;
    FILE* fp=NULL;
    Element* ret;
    for(i=argc-1; 0<i; i--){
        if(argv[i]->type==INT){
            n=argv[i]->integer.val;
        }else if(argv[i]->type==FP){
            fp=argv[i]->file.fp;
        }else if(argv[i]->type==STR){
            type=argv[i]->string.val;
        }
    }
    if(n<1) return NULL;
    if(!fp && !(fp=DefaultFile())){
        CommandError(ToString(argv[0]), E_NoFileSp);
        return NULL;
    }
    if(type==NULL){
        CommandError(ToString(argv[0]), E_NoByteSizeSp);
    }else if(!(size=TypeToSize(type))){
        char msg[256];
        sprintf(msg, E_NoByteSize, type);
        CommandError(ToString(argv[0]), msg);
        return NULL;
    }
    buf=(void*)malloc(n*size);
    if(!fread(buf, size, n, fp)){
        ret=NULL;
    }else if(strcmp(type, "char")==0){
        ret=StringElement((char*)buf);
    }else if(strcmp(type, "int")==0){
        if(n==1){
            ret=IntegerElement(((int*)(buf))[0]);
        }else{
            ret=IArrayElement(n, (int*)buf);
        }
    }else if(strcmp(type, "float")==0){
        if(n==1){
            ret=DoubleElement((double)(((float*)(buf))[0]));
        }else{
            double* a;
            a=(double*)malloc(n*sizeof(double));
            for(i=0; i<n; i++) a[i]=((float*)buf)[i];
            ret=DArrayElement(n, a);
            free(a);
        }
    }else if(strcmp(type, "double")==0){
        if(n==1){
            ret=DoubleElement(((double*)(buf))[0]);
        }else{
            ret=DArrayElement(n, (double*)buf);
        }
    }
    free(buf);
    return ret;
}

Element* C_FSeek(int argc, Element* argv[])
{
    int i, n=1, size=1;
    void* buf;
    char* type=NULL;
    FILE* fp=NULL;
    for(i=argc-1; 0<i; i--){
        if(argv[i]->type==INT){
            n=argv[i]->integer.val;
        }else if(argv[i]->type==FP){
            fp=argv[i]->file.fp;
        }else if(argv[i]->type==STR){
            type=argv[i]->string.val;
        }
    }
    if(!fp && !(fp=DefaultFile())){
        CommandError(ToString(argv[0]), E_NoFileSp);
        return NULL;
    }
    if(type && !(size=TypeToSize(type))){
        char msg[256];
        sprintf(msg, E_NoByteSize, type);
        CommandError(ToString(argv[0]), msg);
        return NULL;
    }
    fseek(fp, (long)(n*size), SEEK_CUR);
    return NULL;
}

Element* C_GetColor(int argc, Element* argv[])
{
    int i, n=0, stat;
    int a[3];
    Element* el=NULL;
    PGColor pgcl;
    if(argc==1){
        CommandError(ToString(argv[0]), E_NoColName);
        return NULL;
    }
    for(i=1; i<argc; i++){
        if(argv[i]->type==STR){
            el=argv[i];
        }else if(argv[i]->type==LIST){
            el=argv[i];
        }else if(argv[i]->type==INT && n<3){
            a[n]=argv[i]->integer.val;  n++;
        }
    }
    if(!el){
        if(3<=n){
            stat=GetColor(a[0], a[1], a[2], &pgcl);
        }else{
            CommandError(ToString(argv[0]), E_BadArgs);
            return NULL;
        }
    }else if(el->type==LIST){
        a[0]=ToInteger(NthElement(1, (List*)argv[1]));
        a[1]=ToInteger(NthElement(2, (List*)argv[1]));
        a[2]=ToInteger(NthElement(3, (List*)argv[1]));
        stat=GetColor(a[0], a[1], a[2], &pgcl);
    }else if(el->type==STR && el->string.val[0]=='#'){
        char buf[3]="ff";
        for(i=0; i<3; i++){
            buf[0]=argv[1]->string.val[2*i+1];
            buf[1]=argv[1]->string.val[2*i+2];
            if(!sscanf(buf, "%x", a+i)){
                char msg[256];
                sprintf(msg, E_BadArg, el->string.val);
                CommandError(ToString(argv[0]), msg);
                return NULL;
            }
        }
        stat=GetColor(a[0], a[1], a[2], &pgcl);
    }else if(el->type==STR){
        stat=GetNamedColor(el->string.val, &pgcl);
    }else{
        CommandError(ToString(argv[0]), E_BadArgs);
        return NULL;
    }
    if(stat) return ColorElement(pgcl);
    CommandError(ToString(argv[0]), E_AllocCol);
    return NULL;
}

Element* C_GetS(int argc, Element* argv[])
{
    int i;
    char buf[256];
    FILE* fp=stdin;
    for(i=argc-1; 0<i; i--){
        if(argv[i]->type==FP) fp=argv[i]->file.fp;
    }
    ReadLine(buf, 256, fp);
    DefaultVariable(V_String);
    return StringElement(buf);
}
Element* C_Global(int argc, Element* argv[])
{
    int i;
    for(i=1; i<argc; i++){
        if(argv[i]->type==VAR) Globalize(argv[i]->string.val);
    }
    return NULL;
}
Element* C_GrayColors(int argc, Element* argv[])
{
    int i, n=17, stat;
    PGColor* pgcl;
    List* ret;
    for(i=argc-1; 0<i; i--){
        if(argv[i]->type==INT) n=argv[i]->integer.val;
    }
    pgcl=(PGColor*)malloc(n*sizeof(PGColor));
    stat=GrayColors(n, pgcl);
    if(!stat){
        CommandError(ToString(argv[0]), E_AllocCol);
        free(pgcl);
        return NULL;
    }
    ret=ToList(NULL);
    for(i=n-1; 0<=i; i--)
            PushToList(ret, ColorElement(pgcl[i]));
    free(pgcl);
    return (Element*)ret;
}

Element* C_IsUndef(int argc, Element* argv[])
{
    if(argc==1) return IntegerElement(TRUE);
    if(argv[1]->type==UNDEF) return IntegerElement(TRUE);
    if(argv[1]->type==VAR){
        Element* el;
        el=VarNameToElement(argv[1]->string.val);
        if(el==NULL || el->type==UNDEF) return IntegerElement(TRUE);
    }
    return IntegerElement(FALSE);
}

Element* C_Limit(int argc, Element* argv[])
{
    Element* elx;
    Element* ely;
    Element* elz;
    elx=C_XLimit(argc, argv);
    ely=C_YLimit(argc, argv);
    elz=C_ZLimit(argc, argv);
    FreeElement(elx);  FreeElement(ely);  FreeElement(elz);
    return NULL;
}
Element* C_LineWidth(int argc, Element* argv[])
{
    int i, n=0;
    double width=0.0;
    if(argc==1){
        printf("%s: %lf\n", ToString(argv[0]), DefaultLineWidth());
        return NULL;
    }
    for(i=1; i<argc; i++){
        if(argv[i]->type==DOUBLE){
            width=argv[i]->real.val;
        }else if(argv[i]->type==INT){
            width=(double)(argv[i]->integer.val);
        }else{
            CommandError(ToString(argv[0]), E_BadArgs);
            return NULL;
        }
    }
    SetDefaultLineWidth(width);
    return NULL;
}

Element* C_Log(int argc, Element* argv[])
{
    int i;
    extern double log();
    List* list;
    list=ToList(NULL);
    if(argc==1) return NULL;
    if(argc==2) return MathElement(argv[1], log);
    for(i=argc-1; 0<i; i--){
        PushToList(list, MathElement(argv[i], log));
    }
    return (Element*)list;
}
Element* C_Log10(int argc, Element* argv[])
{
    int i;
    extern double log10();
    List* list;
    list=ToList(NULL);
    if(argc==1) return NULL;
    if(argc==2) return MathElement(argv[1], log10);
    for(i=argc-1; 0<i; i--){
        PushToList(list, MathElement(argv[i], log10));
    }
    return (Element*)list;
}
Element* C_Maximum(int argc, Element* argv[])
{
    if(argc==1) return NULL;
    return MaxVal(argv[1]);
}
Element* C_MeshLines(int argc, Element* argv[])
{
    int i, n=1, ltype;
    char* str=NULL;
    PGP pgp=GetPGP();
    PTRData* dp=NULL;
    PTRWindow* wp=NULL;
    for(i=argc-1; 0<i; i--){
        if(argv[i]->type==WINDOW){
            wp=argv[i]->window.wp;
        }else if(argv[i]->type==DATA){
            UpdateDataElement((Data*)(argv[i]));
            dp=argv[i]->data.dp;
        }else if(argv[i]->type==STR){
            str=argv[i]->string.val;
        }else if(argv[i]->type==INT){
            n=argv[i]->integer.val;
        }else if(argv[i]->type==COLOR){
            pgp.pgcl=argv[i]->color.pgcl;
        }
    }
    if(str){
        if(!LTNToLT(str, &ltype)){
            char msg[256];
            sprintf(msg, E_NoLineType, str);
            CommandError(ToString(argv[0]), msg);
            return NULL;
        }
    }else{
        ltype=DefaultLineType();
    }
    if(!wp && !(wp=ActiveWindow())){
        CommandError(ToString(argv[0]), E_NoActWin);
        return NULL;
    }
    if(!dp && !(dp=DefaultData())){
        CommandError(ToString(argv[0]), E_NoDataSp);
        return NULL;
    }
    pgp.ltype=ltype;
    DrawMeshLines(wp, dp, n, pgp);
    ShowWindow(wp);
    return NULL;
}
Element* C_Minimum(int argc, Element* argv[])
{
    if(argc==1) return NULL;
    return MinVal(argv[1]);
}
Element* C_MinMax(int argc, Element* argv[])
{
    if(argc==1) return NULL;
    return MinMaxVal(argv[1]);
}

Element* C_NewFrame(int argc, Element* argv[])
{
    Element* ret;
    ret=WindowElement(CreateSimpleFrame());
    DefaultVariable(V_Window);
    return ret;
}

Element* C_NewWindow(int argc, Element* argv[])
{
    int i;
    PTRWindow* wp=NULL;
    PTRWindow* parent=NULL;
    Element* ret;
    wp=ActiveWindow();
    for(i=argc-1; 0<i; i--){
        if(argv[i]->type==WINDOW) parent=argv[i]->window.wp;
    }
    if(!wp){
        parent=CreateSimpleFrame();
    }else if(parent){
    }else if(IsFrameWindow(wp)){
        parent=wp;
    }else{
        parent=ParentWindow(wp);
    }
    ret=WindowElement(CreateSimpleWindow(parent));
    DefaultVariable(V_Window);
    return ret;
}

Element* C_Open(int argc, Element* argv[])
{
    int i;
    char mode[4]="r";
    char* str=NULL;
    char* fname;
    FILE* fp;
    Element* ret;
    for(i=argc-1; 0<i; i--){
        if(argv[i]->type==STR){
            if(IsFileMode(argv[i]->string.val)){
                strcpy(mode, argv[i]->string.val);
            }else{
                str=argv[i]->string.val;
            }
        }
    }
    if(!str){
        CommandError(ToString(argv[0]), E_NoFileSp);
        return NULL;
    }
    fname=ExpandHomeDirectory(str);
    if(!(fp=fopen(fname, mode))){
        char msg[256];
        sprintf(msg, E_FileError, str);
        CommandError(ToString(argv[0]), msg);
        if(fname!=str) free(fname);
        return NULL;
    }
    ret=FileElement(fp, fname, mode);
    if(fname!=str) free(fname);
    DefaultVariable(V_File);
    return ret;
}

Element* C_Plot(int argc, Element* argv[])
{
    int i, n=1, ptype;
    char* str=NULL;
    PGP pgp=GetPGP();
    PTRData* dp=NULL;
    PTRWindow* wp=NULL;
    for(i=argc-1; 0<i; i--){
        if(argv[i]->type==WINDOW){
            wp=argv[i]->window.wp;
        }else if(argv[i]->type==DATA){
            UpdateDataElement((Data*)(argv[i]));
            dp=argv[i]->data.dp;
        }else if(argv[i]->type==STR){
            str=argv[i]->string.val;
        }else if(argv[i]->type==INT){
            n=argv[i]->integer.val;
        }else if(argv[i]->type==COLOR){
            pgp.pgcl=argv[i]->color.pgcl;
        }
    }
    if(str){
        if(!PTNToPT(str, &ptype)){
            char msg[256];
            sprintf(msg, E_NoPointType, str);
            CommandError(ToString(argv[0]), msg);
            return NULL;
        }
    }else{
        ptype=DefaultPointType();
    }
    if(!wp && !(wp=ActiveWindow())){
        CommandError(ToString(argv[0]), E_NoActWin);
        return NULL;
    }
    if(!dp && !(dp=DefaultData())){
        CommandError(ToString(argv[0]), E_NoDataSp);
        return NULL;
    }
    pgp.ptype=ptype;
    DrawPoints(wp, dp, n, pgp);
    ShowWindow(wp);
    return NULL;
}
Element* C_PlotSize(int argc, Element* argv[])
{
    int i, n=0;
    double rad=0.0;
    if(argc==1){
        printf("%s: %lf\n", ToString(argv[0]), DefaultRadius());
        return NULL;
    }
    for(i=1; i<argc; i++){
        if(argv[i]->type==DOUBLE){
            rad=argv[i]->real.val;
        }else if(argv[i]->type==INT){
            rad=(double)(argv[i]->integer.val);
        }else{
            CommandError(ToString(argv[0]), E_BadArgs);
            return NULL;
        }
    }
    SetDefaultRadius(rad);
    return NULL;
}

Element* C_Print(int argc, Element* argv[])
{
    int i;
    int gray=FALSE;
    PTRWindow* wp=NULL;
    for(i=argc-1; 0<i; i--){
        if(strcmp(argv[i]->string.val, "-gray")==0){
            gray=TRUE;
        }else if(argv[i]->type==WINDOW){
            wp=argv[i]->window.wp;
        }
    }
    if(!wp && !(wp=ActiveWindow())){
        CommandError(ToString(argv[0]), E_NoActWin);
        return NULL;
    }
    if(gray) SetPSGrayScale(TRUE);
    PSPrint(ParentFrame(wp));
    if(gray) SetPSGrayScale(FALSE);
    return NULL;
}

Element* C_PrintF(int argc, Element* argv[])
{
    int i=1;
    char* form;  char* str;
    if(argc==1) return NULL;
    form=ToString(argv[i]);
    i++;
    str=SPrintF(form, argc-i, argv+i);
    printf("%s", str);  fflush(stdout);
    if(argv[i-1]->type!=STR) free(form);
    free(str);
    return NULL;
}




Element* C_PSLand(int argc, Element* argv[])
{
    if(argc<2){
        CommandError(ToString(argv[0]), E_NoFileSp);
        return NULL;
    }else{
        PSLandscape(ToString(argv[1]));
    }
    return NULL;
}
Element* C_PSPort(int argc, Element* argv[])
{
    if(argc<2){
        CommandError(ToString(argv[0]), E_NoFileSp);
        return NULL;
    }else{
        PSPortrait(ToString(argv[1]));
    }
    return NULL;
}

Element* C_Quit(int argc, Element* argv[])
{
    SetQuitSignal();
    return NULL;
}

Element* C_Raise(int argc, Element* argv[])
{
    int i;
    PTRWindow* wp=NULL;
    for(i=argc-1; 0<i; i--){
        if(argv[i]->type==WINDOW) wp=argv[i]->window.wp;
    }
    if(!wp && !(wp=ActiveWindow())){
        CommandError(ToString(argv[0]), E_NoActWin);
        return NULL;
    }
    RaiseFrame(wp);
    return NULL;
}

Element* C_Read(int argc, Element* argv[])
{
    int i, n=0, nbuf, line;
    int a[4]={0, 0, 0, 0};
    int xcol=1, ycol=2, col, mcol;
    double* x;
    double* y;
    char buf[256];
    FILE* fp=NULL;
    Element* ret;
    for(i=1; i<argc; i++){
        if(argv[i]->type==INT){
            if(n<4){
                a[n]=argv[i]->integer.val;  n++;
            }
        }else if(argv[i]->type==FP){
            fp=argv[i]->file.fp;
        }
    }
    switch(n){
      case 0:
        break;
      case 1:
        CommandError(ToString(argv[0]), E_BadArgs);
        return NULL;
      case 2:
        xcol=a[0];  ycol=a[1];
    }
    if(xcol<1 || ycol<1){
        CommandError(ToString(argv[0]),
                     "Column number is less than one.\n");
        return NULL;
    }
    if(!fp && !(fp=DefaultFile())){
        CommandError(ToString(argv[0]), E_NoFileSp);
        return NULL;
    }
    fseek(fp, 0L, SEEK_SET);
    if(xcol<ycol){
        mcol=ycol;
    }else{
        mcol=xcol;
    }
    i=0;  nbuf=1024;  line=0;
    x=(double*)malloc(nbuf*sizeof(double));
    y=(double*)malloc(nbuf*sizeof(double));
    while(fgets(buf, 256, fp)){
        List* list;
        line++;
        buf[strlen(buf)-1]='\0';
        StripComments(buf);
        list=Split(' ', buf);
        col=ListLength(list);
        if(col==0){
        }else if(mcol<=col){
            Element* elx;  Element* ely;
            elx=NthElement(xcol, list);
            ely=NthElement(ycol, list);
            if(elx->type!=INT && elx->type!=DOUBLE){
                char msg[256];
                sprintf(msg, E_CannotRead, line, ToString(elx));
                EchoError(msg);
            }else if(elx->type!=INT && elx->type!=DOUBLE){
                char msg[256];
                sprintf(msg, E_CannotRead, line, ToString(ely));
                EchoError(msg);
            }else{
                if(nbuf<=i){
                    nbuf=2*nbuf;
                    x=(double*)realloc(x, nbuf*sizeof(double));
                    y=(double*)realloc(y, nbuf*sizeof(double));
                }
                x[i]=ToDouble(elx);  y[i]=ToDouble(ely);  i++;
            }
        }else{
            char msg[256];
            sprintf(msg, E_ColumnLess, line, mcol);
            EchoError(msg);
        }
        FreeList(list);
    }
    if(i<1){
        /* SetError*/
        return NULL;
    }
    ret=DataElement(CreateSeqData(1, i, 1, x, NULL, NULL, &y));
    free(x);  free(y);
    DefaultVariable(V_Data);
    return ret;
}

Element* C_Redraw(int argc, Element* argv[]){
    int i;
    PTRWindow* wp=NULL;
    for(i=argc-1; 0<i; i--){
        if(argv[i]->type==WINDOW) wp=argv[i]->window.wp;
    }
    if(!wp && !(wp=ActiveWindow())){
        CommandError(ToString(argv[0]), E_NoActWin);
        return NULL;
    }
    ShowWindow(wp);
    return NULL;
}


Element* C_Region(int argc, Element* argv[])
{
    static int rows=1;
    static int cols=1;
    int i, num, n=0;
    int a[3]={0, 0, 0};
    PTRWindow* wp=NULL;
    for(i=1; i<argc; i++){
        if(argv[i]->type==INT
           || argv[i]->type==DOUBLE){
            if(n<3){
                a[n]=ToInteger(argv[i]);  n++;
            }
        }else if(argv[i]->type==WINDOW){
            wp=argv[i]->window.wp;
        }else{
            char msg[256];
            sprintf(msg, E_BadArg, ToString(argv[i]));
            CommandError(ToString(argv[0]), msg);
            return NULL;
        }
    }
    if(n==0){
        CommandError(ToString(argv[0]), E_NoRegNum);
        return NULL;
    }else if(n==1){
        num=a[0];
    }else if(n==2){
        CommandError(ToString(argv[0]), E_FewArgs);
        return NULL;
    }else{
        if(a[0]<1){
            CommandError(ToString(argv[0]), E_LessRows);
            return NULL;
        }else if(a[1]<1){
            CommandError(ToString(argv[0]), E_LessCols);
            return NULL;
        }
        rows=a[0];  cols=a[1];  num=a[2];
    }
    if(!wp && !(wp=ActiveWindow())){
        CommandError(ToString(argv[0]), E_NoActWin);
        return NULL;
    }
    WindowRegion(wp, rows, cols, num);
    return NULL;
}

Element* C_ResizeFrame(int argc, Element* argv[])
{
    int i, n=0;
    int a[2]={0, 0};
    PTRWindow* wp=NULL;
    for(i=1; i<argc; i++){
        if(argv[i]->type==WINDOW){
            wp=argv[i]->window.wp;
        }else if(argv[i]->type==INT
                 || argv[i]->type==DOUBLE){
            if(n<2){
                a[n]=ToInteger(argv[i]);  n++;
            }
        }
    }
    if(!wp && !(wp=ActiveWindow())){
        CommandError(ToString(argv[0]), E_NoActWin);
        return NULL;
    }
    if(n<1){
        CommandError(ToString(argv[0]), E_FewArgs);
        return NULL;
    }
    ResizeFrame(wp, (unsigned int)a[0], (unsigned int)a[1]);
    return NULL;
}

Element* C_RestoreData(int argc, Element* argv[])
{
    int i;
    char* str=NULL;
    FILE* fp=NULL;
    Element* ret;
    for(i=argc-1; 0<i; i--){
        if(argv[i]->type==FP){
            fp=argv[i]->file.fp;
        }else if(argv[i]->type==STR){
            str=argv[i]->string.val;
        }
    }
    if(str){
        ret=DataElement(RestoreData(str));
    }else if(!fp && !(fp=DefaultFile())){
        CommandError(ToString(argv[0]), E_NoFileSp);
        return NULL;
    }else{
        ret=DataElement(RestoreDataFP(fp));
    }
    DefaultVariable(V_Data);
    return ret;
}
Element* C_Return(int argc, Element* argv[])
{
    SetReturnSignal();
    if(argc==1) return NULL;
    return argv[1];
}

Element* C_SaveData(int argc, Element* argv[])
{
    int i;
    char* str=NULL;
    FILE* fp=NULL;
    PTRData* dp=NULL;
    for(i=argc-1; 0<i; i--){
        if(argv[i]->type==FP){
            fp=argv[i]->file.fp;
        }else if(argv[i]->type==STR){
            str=argv[i]->string.val;
        }else if(argv[i]->type==DATA){
            UpdateDataElement((Data*)(argv[i]));
            dp=argv[i]->data.dp;
        }
    }
    if(!dp && !(dp=DefaultData())){
        CommandError(ToString(argv[0]), E_NoData);
        return NULL;
    }
    if(str){
        SaveData(str, dp);
    }else if(!fp && !(fp=DefaultFile())){
        CommandError(ToString(argv[0]), E_NoFileSp);
        return NULL;
    }else{
        SaveDataFP(fp, dp);
    }
    return NULL;
}

Element* C_Scale(int argc, Element* argv[])
{
    int i, set=0, ax=AX_NL;
    char* form=NULL;
    PTRWindow* wp=NULL;
    Element* el=NULL;
    for(i=1; i<argc; i++){
        if(argv[i]->type==WINDOW){
            wp=argv[i]->window.wp;
        }else if(argv[i]->type==STR){
            if(strcmp(argv[i]->string.val, "-form")==0){
                i++;
                if(argc<=i){
                    CommandError(ToString(argv[0]), E_BadArgs);
                    return NULL;
                }
                form=ToString(argv[i]);
                set=1;
            }else if(strcmp(argv[i]->string.val, "-ax")==0){
                i++;
                if(argc<=i){
                    CommandError(ToString(argv[0]), E_BadArgs);
                    return NULL;
                }
                ax=NameToAxis(argv[i]->string.val);
                if(ax<0 || 4<ax){
                     char msg[256];
                     sprintf(msg, E_NoSuchAxis, ToString(argv[i]));
                     CommandError(ToString(argv[0]), msg);
                     return NULL;
                }
                set=1;
            }else{
                el=argv[i];
            }
        }else{
            el=argv[i];
        }
    }
    if(!wp && !(wp=ActiveWindow())){
        CommandError(ToString(argv[0]), E_NoActWin);
        return NULL;
    }
    if(set!=0){
    }else if(el){
        set=ToBoolean(el)-HasScale(wp);
    }else{
        set=1-2*HasScale(wp);
    }
    if(set==1){
        if(1<=ax && ax<=4){
            AutoAxScale(wp, ax, form);
        }else{
            SetScale(wp, form);
        }
    }else if(set==-1){
        RemoveScale(wp);
    }else{
        return NULL;
    }
    ShowWindow(wp);
    return NULL;
}

Element* C_ShowPage(int argc, Element* argv[]){
    ShowPage();
    return NULL;
}

Element* C_Sin(int argc, Element* argv[])
{
    int i;
    extern double sin();
    List* list;
    list=ToList(NULL);
    if(argc==1) return NULL;
    if(argc==2) return MathElement(argv[1], sin);
    for(i=argc-1; 0<i; i--){
        PushToList(list, MathElement(argv[i], sin));
    }
    return (Element*)list;
}

Element* C_Split(int argc, Element* argv[])
{
    int i=1, nbuf=0;
    char c=' ';
    char* buf=NULL;
    List* list=NULL;
    if(2<argc && argv[1]->type==STR
       && argv[1]->string.val[1]=='\0'){
        c=argv[1]->string.val[0];
        i++;
    }
    if(argc<=i){
        char* str;
        if(!(str=DefaultString())){
            CommandError(ToString(argv[0]), E_NoStrSp);
            return NULL;
        }
        buf=(char*)malloc((strlen(str)+1)*sizeof(char));
        strcpy(buf, str);
        list=Split(c, buf);
    }
    for(; i<argc; i++){
        if(argv[i]->type==STR){
            int l=strlen(argv[i]->string.val);
            if(nbuf<l+1){
                if(buf) free(buf);
                nbuf=l+1;
                buf=(char*)malloc(nbuf*sizeof(char));
            }
            strcpy(buf, argv[i]->string.val);
            list=CatList(list, Split(c, buf));
        }
    }
    if(buf) free(buf);
    return (Element*)list;
}
Element* C_SPrintF(int argc, Element* argv[])
{
    int i=1;
    char* form;  char* str;
    Element* ret;
    if(argc==1) return NULL;
    form=ToString(argv[i]);
    i++;
    str=SPrintF(form, argc-i, argv+i);
    ret=StringElement(str);
    if(argv[i-1]->type!=STR) free(form);
    free(str);
    return ret;
}


Element* C_Sqrt(int argc, Element* argv[])
{
    int i;
    extern double sqrt();
    List* list;
    list=ToList(NULL);
    if(argc==1) return NULL;
    if(argc==2) return MathElement(argv[1], sqrt);
    for(i=argc-1; 0<i; i--){
        PushToList(list, MathElement(argv[i], sqrt));
    }
    return (Element*)list;
}

Element* C_System(int argc, Element* argv[])
{
    int i, nbuf=256;
    char* str;
    char* s;
    str=(char*)malloc(nbuf*sizeof(char));
    for(i=1; i<argc; i++){
        s=ToString(argv[i]);
        if(nbuf<=strlen(str)+strlen(s)+1){
            nbuf+=256;
            str=(char*)realloc(str, nbuf*sizeof(char));
        }
        strcat(str, s);
        strcat(str, " ");
        if(argv[i]->type!=STR) free(s);
    }
    system(str);
    free(str);
    return NULL;
}

Element* C_Tan(int argc, Element* argv[])
{
    int i;
    extern double tan();
    List* list;
    list=ToList(NULL);
    if(argc==1) return NULL;
    if(argc==2) return MathElement(argv[1], tan);
    for(i=argc-1; 0<i; i--){
        PushToList(list, MathElement(argv[i], tan));
    }
    return (Element*)list;
}

Element* C_ToInt(int argc, Element* argv[])
{
    if(argc==1) return NULL;
    return IntegerElement(ToInteger(argv[1]));
}

Element* C_TopLabel(int argc, Element* argv[])
{
    int i;
    char* str=NULL;
    PTRWindow* wp=NULL;
    for(i=argc-1; 0<i; i--){
        if(argv[i]->type==WINDOW){
            wp=argv[i]->window.wp;
        }else if(argv[i]->type==STR){
            str=argv[i]->string.val;
        }
    }
    if(!wp && !(wp=ActiveWindow())){
        CommandError(ToString(argv[0]), E_NoActWin);
        return NULL;
    }
    if(!str){
        printf("%s: %s\n", ToString(argv[0]), TopLabelString(wp));
        return NULL;
    }
    if(str[0]=='\0'){
        RemoveTopLabel(wp);
    }else{
        SetTopLabel(wp, str, GetPGP());
    }
    ShowWindow(wp);
    return NULL;
}

Element* C_ToGif(int argc, Element* argv[])
{
    int i;
    char* str=NULL;
    PTRWindow* wp=NULL;
    for(i=argc-1; 0<i; i--){
        if(argv[i]->type==WINDOW){
            wp=argv[i]->window.wp;
        }else if(argv[i]->type==STR){
            str=argv[i]->string.val;
        }
    }
    if(str==NULL){
        CommandError(ToString(argv[0]), E_NoFileSp);
        return NULL;
    }
    if(!wp && !(wp=ActiveWindow())){
        CommandError(ToString(argv[0]), E_NoActWin);
        return NULL;
    }
    DumpPTRWindowToGif(wp, str);
    return NULL;
}
Element* C_ToXpm(int argc, Element* argv[])
{
    int i;
    char* str=NULL;
    PTRWindow* wp=NULL;
    for(i=argc-1; 0<i; i--){
        if(argv[i]->type==WINDOW){
            wp=argv[i]->window.wp;
        }else if(argv[i]->type==STR){
            str=argv[i]->string.val;
        }
    }
    if(str==NULL){
        CommandError(ToString(argv[0]), E_NoFileSp);
        return NULL;
    }
    if(!wp && !(wp=ActiveWindow())){
        CommandError(ToString(argv[0]), E_NoActWin);
        return NULL;
    }
    DumpPTRWindowToXpm(wp, str);
    return NULL;
}

Element* C_Version(int argc, Element* argv[])
{
    EchoVersion();
    return NULL;
}

Element* C_WaitKey(int argc, Element* argv[])
{
    char key[2]="\0\0";
    WaitKeyPress(key, NULL, NULL);
    return StringElement(key);
};

Element* C_VLabel(int argc, Element* argv[])
{
    int i, rot=R0;
    char* str=NULL;
    PGP pgp=GetPGP();
    PTRWindow* wp=NULL;
    for(i=argc-1; 0<i; i--){
        if(argv[i]->type==WINDOW){
            wp=argv[i]->window.wp;
        }else if(argv[i]->type==STR){
            str=argv[i]->string.val;
        }else if(argv[i]->type==INT || argv[i]->type==DOUBLE){
            rot=ToInteger(argv[i]);
        }
    }
    if(!wp && !(wp=ActiveWindow())){
        CommandError(ToString(argv[0]), E_NoActWin);
        return NULL;
    }
    if(!str){
        printf("%s: %s\n", ToString(argv[0]), VLabelString(wp));
        return NULL;
    }
    if(str[0]=='\0'){
        RemoveVLabel(wp);
    }else{
        pgp.rot=rot;
        SetVLabel(wp, str, pgp);
    }
    ShowWindow(wp);
    return NULL;
}

Element* C_VLimit(int argc, Element* argv[])
{
    int i, n=0;
    double a[2];
    PTRData* dp=NULL;
    PTRWindow* wp=NULL;
    for(i=1; i<argc; i++){
        if(argv[i]->type==WINDOW){
            wp=argv[i]->window.wp;
        }else if(argv[i]->type==DATA){
            UpdateDataElement((Data*)(argv[i]));
            dp=argv[i]->data.dp;
        }else if(argv[i]->type==INT || argv[i]->type==DOUBLE){
            if(n<2){
                a[n]=ToDouble(argv[i]);  n++;
            }
        }
    }
    if(!wp && !(wp=ActiveWindow())){
        CommandError(ToString(argv[0]), E_NoActWin);
        return NULL;
    }
    if(n==0){
        double min, max;
        if(!dp && !(dp=DefaultData())){
            CommandError(ToString(argv[0]), E_NoDataSp);
            return NULL;
        }
        if(!ExtractVRange(dp, 1, &min, &max)) return NULL;
        if(min<max){
            LimitExpand(&min, &max);
            SetVLimit(wp, min, max);
        }else{
            char msg[256];
            sprintf(msg, E_SameValue, min);
            CommandError(ToString(argv[0]), msg);
            return NULL;
        }
    }else if(n==2){
        SetVLimit(wp, a[0], a[1]);
    }else{
        CommandError(ToString(argv[0]), E_FewArgs);
        return NULL;
    }
    ShowWindow(wp);
    return NULL;
}


Element* C_X1Label(int argc, Element* argv[])
{
    int i;
    char* str=NULL;
    PTRWindow* wp=NULL;
    for(i=argc-1; 0<i; i--){
        if(argv[i]->type==WINDOW){
            wp=argv[i]->window.wp;
        }else if(argv[i]->type==STR){
            str=argv[i]->string.val;
        }
    }
    if(!wp && !(wp=ActiveWindow())){
        CommandError(ToString(argv[0]), E_NoActWin);
        return NULL;
    }
    if(!str){
        printf("%s: %s\n", ToString(argv[0]), X1LabelString(wp));
        return NULL;
    }
    if(str[0]=='\0'){
        RemoveX1Label(wp);
    }else{
        SetX1Label(wp, str, GetPGP());
    }
    ShowWindow(wp);
    return NULL;
}

Element* C_X1Limit(int argc, Element* argv[])
{
    int i, n=0;
    double a[2];
    PTRData* dp=NULL;
    PTRWindow* wp=NULL;
    for(i=1; i<argc; i++){
        if(argv[i]->type==WINDOW){
            wp=argv[i]->window.wp;
        }else if(argv[i]->type==DATA){
            UpdateDataElement((Data*)(argv[i]));
            dp=argv[i]->data.dp;
        }else if(argv[i]->type==INT || argv[i]->type==DOUBLE){
            if(n<2){
                a[n]=ToDouble(argv[i]);  n++;
            }
        }
    }
    if(!wp && !(wp=ActiveWindow())){
        CommandError(ToString(argv[0]), E_NoActWin);
        return NULL;
    }
    if(n==0){
        double min, max, eps=1.0e-16;
        if(!dp && !(dp=DefaultData())){
            CommandError(ToString(argv[0]), E_NoDataSp);
            return NULL;
        }
        if(!ExtractX1Range(dp, &min, &max)) return NULL;
        if(eps<max-min){
            LimitExpand(&min, &max);
            SetX1Limit(wp, min, max);
        }else{
            char msg[256];
            sprintf(msg, E_SameValue, min);
            CommandError(ToString(argv[0]), msg);
            return NULL;
        }
    }else if(n==2){
        SetX1Limit(wp, a[0], a[1]);
    }else{
        CommandError(ToString(argv[0]), E_FewArgs);
        return NULL;
    }
    ShowWindow(wp);
    return NULL;
}


Element* C_X2Label(int argc, Element* argv[])
{
    int i, rot=R0;
    char* str=NULL;
    PGP pgp=GetPGP();
    PTRWindow* wp=NULL;
    for(i=argc-1; 0<i; i--){
        if(argv[i]->type==WINDOW){
            wp=argv[i]->window.wp;
        }else if(argv[i]->type==STR){
            str=argv[i]->string.val;
        }else if(argv[i]->type==INT || argv[i]->type==DOUBLE){
            rot=ToInteger(argv[i]);
        }
    }
    if(!wp && !(wp=ActiveWindow())){
        CommandError(ToString(argv[0]), E_NoActWin);
        return NULL;
    }
    if(!str){
        printf("%s: %s\n", ToString(argv[0]), X2LabelString(wp));
        return NULL;
    }
    if(str[0]=='\0'){
        RemoveX2Label(wp);
    }else{
        pgp.rot=rot;
        SetX2Label(wp, str, pgp);
    }
    ShowWindow(wp);
    return NULL;
}

Element* C_X2Limit(int argc, Element* argv[])
{
    int i, n=0;
    double a[2];
    PTRData* dp=NULL;
    PTRWindow* wp=NULL;
    for(i=1; i<argc; i++){
        if(argv[i]->type==WINDOW){
            wp=argv[i]->window.wp;
        }else if(argv[i]->type==DATA){
            UpdateDataElement((Data*)(argv[i]));
            dp=argv[i]->data.dp;
        }else if(argv[i]->type==INT || argv[i]->type==DOUBLE){
            if(n<2){
                a[n]=ToDouble(argv[i]);  n++;
            }
        }
    }
    if(!wp && !(wp=ActiveWindow())){
        CommandError(ToString(argv[0]), E_NoActWin);
        return NULL;
    }
    if(n==0){
        double min, max, eps=1.0e-16;
        if(!dp && !(dp=DefaultData())){
            CommandError(ToString(argv[0]), E_NoDataSp);
            return NULL;
        }
        if(!ExtractX2Range(dp, &min, &max)) return NULL;
        if(eps<max-min){
            LimitExpand(&min, &max);
            SetX2Limit(wp, min, max);
        }else{
            char msg[256];
            sprintf(msg, E_SameValue, min);
            CommandError(ToString(argv[0]), msg);
            return NULL;
        }
    }else if(n==2){
        SetX2Limit(wp, a[0], a[1]);
    }else{
        CommandError(ToString(argv[0]), E_FewArgs);
        return NULL;
    }
    ShowWindow(wp);
    return NULL;
}


Element* C_X3Label(int argc, Element* argv[])
{
    int i, rot=R0;
    char* str=NULL;
    PGP pgp=GetPGP();
    PTRWindow* wp=NULL;
    for(i=argc-1; 0<i; i--){
        if(argv[i]->type==WINDOW){
            wp=argv[i]->window.wp;
        }else if(argv[i]->type==STR){
            str=argv[i]->string.val;
        }else if(argv[i]->type==INT || argv[i]->type==DOUBLE){
            rot=ToInteger(argv[i]);
        }
    }
    if(!wp && !(wp=ActiveWindow())){
        CommandError(ToString(argv[0]), E_NoActWin);
        return NULL;
    }
    if(!str){
        printf("%s: %s\n", ToString(argv[0]), X3LabelString(wp));
        return NULL;
    }
    if(str[0]=='\0'){
        RemoveX3Label(wp);
    }else{
        pgp.rot=rot;
        SetX3Label(wp, str, pgp);
    }
    ShowWindow(wp);
    return NULL;
}

Element* C_X3Limit(int argc, Element* argv[])
{
    int i, n=0;
    double a[2];
    PTRData* dp=NULL;
    PTRWindow* wp=NULL;
    for(i=1; i<argc; i++){
        if(argv[i]->type==WINDOW){
            wp=argv[i]->window.wp;
        }else if(argv[i]->type==DATA){
            UpdateDataElement((Data*)(argv[i]));
            dp=argv[i]->data.dp;
        }else if(argv[i]->type==INT || argv[i]->type==DOUBLE){
            if(n<2){
                a[n]=ToDouble(argv[i]);  n++;
            }
        }
    }
    if(!wp && !(wp=ActiveWindow())){
        CommandError(ToString(argv[0]), E_NoActWin);
        return NULL;
    }
    if(n==0){
        double min, max, eps=1.0e-16;
        if(!dp && !(dp=DefaultData())){
            CommandError(ToString(argv[0]), E_NoDataSp);
            return NULL;
        }
        if(!ExtractX3Range(dp, &min, &max)) return NULL;
        if(eps<max-min){
            LimitExpand(&min, &max);
            SetX3Limit(wp, min, max);
        }else{
            char msg[256];
            sprintf(msg, E_SameValue, min);
            CommandError(ToString(argv[0]), msg);
            return NULL;
        }
    }else if(n==2){
        SetX3Limit(wp, a[0], a[1]);
    }else{
        CommandError(ToString(argv[0]), E_FewArgs);
        return NULL;
    }
    ShowWindow(wp);
    return NULL;
}
Element* C_XLabel(int argc, Element* argv[])
{
    int i;
    char* str=NULL;
    PTRWindow* wp=NULL;
    for(i=argc-1; 0<i; i--){
        if(argv[i]->type==WINDOW){
            wp=argv[i]->window.wp;
        }else if(argv[i]->type==STR){
            str=argv[i]->string.val;
        }
    }
    if(!wp && !(wp=ActiveWindow())){
        CommandError(ToString(argv[0]), E_NoActWin);
        return NULL;
    }
    if(!str){
        printf("%s: %s\n", ToString(argv[0]), XLabelString(wp));
        return NULL;
    }
    if(str[0]=='\0'){
        RemoveXLabel(wp);
    }else{
        SetXLabel(wp, str, GetPGP());
    }
    ShowWindow(wp);
    return NULL;
}
Element* C_XLimit(int argc, Element* argv[])
{
    int i, n=0;
    double a[2];
    PTRData* dp=NULL;
    PTRWindow* wp=NULL;
    for(i=1; i<argc; i++){
        if(argv[i]->type==WINDOW){
            wp=argv[i]->window.wp;
        }else if(argv[i]->type==DATA){
            UpdateDataElement((Data*)(argv[i]));
            dp=argv[i]->data.dp;
        }else if(argv[i]->type==INT || argv[i]->type==DOUBLE){
            if(n<2){
                a[n]=ToDouble(argv[i]);  n++;
            }
        }
    }
    if(!wp && !(wp=ActiveWindow())){
        CommandError(ToString(argv[0]), E_NoActWin);
        return NULL;
    }
    if(n==0){
        double min, max, eps=1.0e-16;
        if(!dp && !(dp=DefaultData())){
            CommandError(ToString(argv[0]), E_NoDataSp);
            return NULL;
        }
        if(!ExtractXRange(wp, dp, 1, &min, &max)) return NULL;
        if(eps<max-min){
            LimitExpand(&min, &max);
            SetXLimit(wp, min, max);
        }else{
            char msg[256];
            sprintf(msg, E_SameValue, min);
            CommandError(ToString(argv[0]), msg);
            return NULL;
        }
    }else if(n==2){
        SetXLimit(wp, a[0], a[1]);
    }else{
        CommandError(ToString(argv[0]), E_FewArgs);
        return NULL;
    }
    ShowWindow(wp);
    return NULL;
}
Element* C_YLabel(int argc, Element* argv[])
{
    int i, rot=R0;
    char* str=NULL;
    PGP pgp=GetPGP();
    PTRWindow* wp=NULL;
    for(i=argc-1; 0<i; i--){
        if(argv[i]->type==WINDOW){
            wp=argv[i]->window.wp;
        }else if(argv[i]->type==STR){
            str=argv[i]->string.val;
        }else if(argv[i]->type==INT || argv[i]->type==DOUBLE){
            rot=ToInteger(argv[i]);
        }
    }
    if(!wp && !(wp=ActiveWindow())){
        CommandError(ToString(argv[0]), E_NoActWin);
        return NULL;
    }
    if(!str){
        printf("%s: %s\n", ToString(argv[0]), YLabelString(wp));
        return NULL;
    }
    if(str[0]=='\0'){
        RemoveYLabel(wp);
    }else{
        pgp.rot=rot;
        SetYLabel(wp, str, pgp);
    }
    ShowWindow(wp);
    return NULL;
}
Element* C_YLimit(int argc, Element* argv[])
{
    int i, n=0;
    double a[2];
    PTRData* dp=NULL;
    PTRWindow* wp=NULL;
    for(i=1; i<argc; i++){
        if(argv[i]->type==WINDOW){
            wp=argv[i]->window.wp;
        }else if(argv[i]->type==DATA){
            UpdateDataElement((Data*)(argv[i]));
            dp=argv[i]->data.dp;
        }else if(argv[i]->type==INT || argv[i]->type==DOUBLE){
            if(n<2){
                a[n]=ToDouble(argv[i]);  n++;
            }
        }
    }
    if(!wp && !(wp=ActiveWindow())){
        CommandError(ToString(argv[0]), E_NoActWin);
        return NULL;
    }
    if(n==0){
        double min, max, eps=1.0e-16;
        if(!dp && !(dp=DefaultData())){
            CommandError(ToString(argv[0]), E_NoDataSp);
            return NULL;
        }
        if(!ExtractYRange(wp, dp, 1, &min, &max)) return NULL;
        if(eps<max-min){
            LimitExpand(&min, &max);
            SetYLimit(wp, min, max);
        }else{
            char msg[256];
            sprintf(msg, E_SameValue, min);
            CommandError(ToString(argv[0]), msg);
            return NULL;
        }
    }else if(n==2){
        SetYLimit(wp, a[0], a[1]);
    }else{
        CommandError(ToString(argv[0]), E_FewArgs);
        return NULL;
    }
    ShowWindow(wp);
    return NULL;
}
Element* C_ZLabel(int argc, Element* argv[])
{
    int i, rot=R0;
    char* str=NULL;
    PGP pgp=GetPGP();
    PTRWindow* wp=NULL;
    for(i=argc-1; 0<i; i--){
        if(argv[i]->type==WINDOW){
            wp=argv[i]->window.wp;
        }else if(argv[i]->type==STR){
            str=argv[i]->string.val;
        }else if(argv[i]->type==INT || argv[i]->type==DOUBLE){
            rot=ToInteger(argv[i]);
        }
    }
    if(!wp && !(wp=ActiveWindow())){
        CommandError(ToString(argv[0]), E_NoActWin);
        return NULL;
    }
    if(!str){
        printf("%s: %s\n", ToString(argv[0]), ZLabelString(wp));
        return NULL;
    }
    if(str[0]=='\0'){
        RemoveZLabel(wp);
    }else{
        pgp.rot=rot;
        SetZLabel(wp, str, pgp);
    }
    ShowWindow(wp);
    return NULL;
}
Element* C_ZLimit(int argc, Element* argv[])
{
    int i, n=0;
    double a[2];
    PTRData* dp=NULL;
    PTRWindow* wp=NULL;
    for(i=1; i<argc; i++){
        if(argv[i]->type==WINDOW){
            wp=argv[i]->window.wp;
        }else if(argv[i]->type==DATA){
            UpdateDataElement((Data*)(argv[i]));
            dp=argv[i]->data.dp;
        }else if(argv[i]->type==INT || argv[i]->type==DOUBLE){
            if(n<2){
                a[n]=ToDouble(argv[i]);  n++;
            }
        }
    }
    if(!wp && !(wp=ActiveWindow())){
        CommandError(ToString(argv[0]), E_NoActWin);
        return NULL;
    }
    if(n==0){
        double min, max, eps=1.0e-16;
        if(!dp && !(dp=DefaultData())){
            CommandError(ToString(argv[0]), E_NoDataSp);
            return NULL;
        }
        if(!ExtractZRange(wp, dp, 1, &min, &max)) return NULL;
        if(eps<max-min){
            LimitExpand(&min, &max);
            SetZLimit(wp, min, max);
        }else{
            char msg[256];
            sprintf(msg, E_SameValue, min);
            CommandError(ToString(argv[0]), msg);
            return NULL;
        }
    }else if(n==2){
        SetZLimit(wp, a[0], a[1]);
    }else{
        CommandError(ToString(argv[0]), E_FewArgs);
        return NULL;
    }
    ShowWindow(wp);
    return NULL;
}
