#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <pwd.h>
#include <ptr.h>
#include "pointer.h"


TypeList pointtypes[]=
{
    {"cir", CIR},
    {"circle", CIR},
    {"dot", PTS},
    {"fcir", FCIR},
    {"fcircle", FCIR},
    {"fsqr", FSQR},
    {"fsquare", FSQR},
    {"points", PTS},
    {"pts", PTS},
    {"sqr", SQR},
    {"square", SQR}
};
int npointtypes=sizeof(pointtypes)/sizeof(*pointtypes);
TypeList linetypes[]=
{
    {"brk", BRK},
    {"broken", BRK},
    {"dot", DOT},
    {"sol", SOL},
    {"solid", SOL}
};
int nlinetypes=sizeof(linetypes)/sizeof(*linetypes);
TypeList fonts[]=
{
    {"courier", COUR},
    {"times", TIMS},
    {"helvetica", HELV},
    {"symbol", SYMB}
};
int nfonts=sizeof(fonts)/sizeof(*fonts);
TypeList slants[]=
{
    {"roman", RM},
    {"italic", IT}
};
int nslants=sizeof(slants)/sizeof(*slants);
TypeList bytesizes[]=
{
    {"char", sizeof(char)},
    {"int", sizeof(int)},
    {"float", sizeof(float)},
    {"double", sizeof(double)}
};
int nbytesizes=sizeof(bytesizes)/sizeof(*bytesizes);
TypeList types[]=
{
    {"undef", UNDEF},
    {"int", INT},
    {"double", DOUBLE},
    {"string", STR},
    {"file", FP},
    {"data", DATA},
    {"window", WINDOW},
    {"integer array", IARRAY},
    {"double array", DARRAY},
    {"list", LIST}
};
int ntypes=sizeof(types)/sizeof(*types);

int e_parse=FALSE;
int firsterr=TRUE;
char filename[256]="";
char subname[256]="";

int s_quit =FALSE;
int s_break=FALSE;
int s_cont =FALSE;
int s_ret  =FALSE;

void SetQuitSignal(){
    s_quit=TRUE;
}
int QuitSignal(){
    return s_quit;
}
void ClearQuitSignal(){
    s_quit=FALSE;
}
void SetBreakSignal(){
    s_break=TRUE;
}
int BreakSignal(){
    return s_break;
}
void ClearBreakSignal(){
    s_break=FALSE;
}
void SetContinueSignal(){
    s_cont=TRUE;
}
int ContinueSignal(){
    return s_cont;
}
void ClearContinueSignal(){
    s_cont=FALSE;
}
void SetReturnSignal(){
    s_ret=TRUE;
}
int ReturnSignal(){
    return s_ret;
}
void ClearReturnSignal(){
    s_ret=FALSE;
}
int AnySignal(){
    return (s_quit || s_break || s_cont || s_ret);
}

int PTNToPT(char* ptname, int* ptype)
{
    TypeList* tp;
    for(tp=pointtypes; tp<pointtypes+npointtypes; tp++){
        if(strcmp(ptname, tp->name)) continue;
        *ptype=tp->type;
        return TRUE;
    }
    return FALSE;
}
int LTNToLT(char* ltname, int* ltype)
{
    TypeList* tp;
    for(tp=linetypes; tp<linetypes+nlinetypes; tp++){
        if(strcmp(ltname, tp->name)) continue;
        *ltype=tp->type;
        return TRUE;
    }
    return FALSE;
}
int FontNameToFont(char* ftname, int* font)
{
    TypeList* tp;
    for(tp=fonts; tp<fonts+nfonts; tp++){
        if(strcmp(ftname, tp->name)) continue;
        *font=tp->type;
        return TRUE;
    }
    return FALSE;
}
char* FontToFontName(int font)
{
    TypeList* tp;
    for(tp=fonts; tp<fonts+nfonts; tp++){
        if(font!=tp->type) continue;
        return tp->name;
    }
    return "";
}
int SlantNameToSlant(char* slname, int* slant)
{
    TypeList* tp;
    for(tp=slants; tp<slants+nslants; tp++){
        if(strcmp(slname, tp->name)) continue;
        *slant=tp->type;
        return TRUE;
    }
    return FALSE;
}
char* SlantToSlantName(int slant)
{
    TypeList* tp;
    for(tp=slants; tp<slants+nslants; tp++){
        if(slant!=tp->type) continue;
        return tp->name;
    }
    return "";
}
int TypeToSize(char* type)
{
    TypeList* tp;
    for(tp=bytesizes; tp<bytesizes+nbytesizes; tp++){
        if(strcmp(type, tp->name)) continue;
        return tp->type;
    }
    return 0;
}
char* TypeToName(int type)
{
    TypeList* tp;
    for(tp=types; tp<types+ntypes; tp++){
        if(type!=tp->type) continue;
        return tp->name;
    }
    return "";
}
char* DataTypeToName(int type){
    if(type==GRID){
        return "grid";
    }else if(type==SEQ){
        return "sequential";
    }else{
        return "";
    }
}
int NameToDataType(char* type){
    if(strcmp(type, "grid")==0){
        return GRID;
    }else if(strcmp(type, "sequential")==0
             || strcmp(type, "seq")==0){
        return SEQ;
    }else{
        return 0;
    }
}
char* AxisToName(int ax){
    if(ax==AX_X1){
        return "x1";
    }else if(ax==AX_X2){
        return "x2";
    }else if(ax==AX_X3){
        return "x3";
    }else if(ax==AX_V){
        return "v";
    }else{
        return "none";
    }
}
int NameToAxis(char* ax){
    if(strcmp(ax, "x1")==0 || strcmp(ax, "x")==0){
        return AX_X1;
    }else if(strcmp(ax, "x2")==0 || strcmp(ax, "y")==0){
        return AX_X2;
    }else if(strcmp(ax, "x3")==0 || strcmp(ax, "z")==0){
        return AX_X3;
    }else if(strcmp(ax, "v")==0){
        return AX_V;
    }else if(strcmp(ax, "none")==0){
        return AX_NL;
    }else{
        return -1;
    }
}
char* CoordToName(int coord){
    if(coord==CAR){
        return "Cartesian";
    }else if(coord==CYL){
        return "cylindrical";
    }else{
        return "";
    }
}
int NameToCoord(char* coord){
    if(strcmp(coord, "Cartesian")==0
       || strcmp(coord, "cart")==0){
        return CAR;
    }else if(strcmp(coord, "cylindrical")==0
             || strcmp(coord, "cyl")==0){
        return CYL;
    }else{
        return 0;
    }
}
int IsWord(char c){
    if(isalnum(c) || c=='_') return TRUE;
    return FALSE;
}
int IsVarWord(char c){
    if(IsWord(c) || c=='$') return TRUE;
    return FALSE;
}
int IsQuote(char c){
    if(c=='\'' || c=='\"') return TRUE;
    return FALSE;
}
int IsLeftParentheses(char c){
    if(c=='(' || c=='{' || c=='[') return TRUE;
    return FALSE;
}
int IsRightParentheses(char c){
    if(c==')' || c=='}' || c==']') return TRUE;
    return FALSE;
}
char OtherParentheses(char c){
    switch(c){
      case '\'':
        return '\'';
        break;
      case '\"':
        return '\"';
        break;
      case '(':
        return ')';
        break;
      case '{':
        return '}';
        break;
      case '[':
        return ']';
        break;
      case ')':
        return '(';
        break;
      case '}':
        return '{';
        break;
      case ']':
        return '[';
        break;
      default:
        return 0;
    }
}
int IsFileMode(char* str){
    switch(str[0]){
      case 'r':
      case 'w':
      case 'a':
        break;
      default:
        return FALSE;
    }
    switch(str[1]){
      case '\0':
        return TRUE;
        break;
      case '+':
        break;
      default:
        return FALSE;
    }
    if(str[2]=='\0') return TRUE;
    return FALSE;
}

int IsArray(Element* el){
    if(el==NULL) return FALSE;
    return (el->type==IARRAY || el->type==DARRAY);
}
int IsNode(Element* el){
    if(el==NULL) return FALSE;
    if(N_Node<=el->type) return TRUE;
    return FALSE;
}
int IsFunctionForm(char* str, int i){
    if(!(isalpha(str[0]) || str[0]=='\\')) return FALSE;
    for(; 0<i; i--){
        if(!IsWord(str[i])) return FALSE;
    }
    return TRUE;
}
int IsCommandForm(char* str){
    int i=0;
    if(str[i]=='\\') i++;
    if(!isalpha(str[i])) return FALSE;
    while(IsWord(str[i])) i++;
    if(str[i]==' ' || str[i]=='\0') return TRUE;
    return FALSE;
}
int IsIfBlock(char* str){
    if(str[0]=='i' && str[1]=='f'
       && (str[2]=='(' || str[2]=='{' || str[2]==' ')) return TRUE;
    return FALSE;
}
int IsElseBlock(char* str){
    if(str[0]=='e' && str[1]=='l'
       && str[2]=='s' && str[3]=='e') return TRUE;
    return FALSE;
}
int IsWhileBlock(char* str){
    if(str[0]=='w' && str[1]=='h' && str[2]=='i'
       && str[3]=='l' && str[4]=='e'
       && (str[5]=='(' || str[5]=='{' || str[5]==' ')) return TRUE;
    return FALSE;
}
int IsForBlock(char* str){
    if(str[0]=='f' && str[1]=='o' && str[2]=='r'
       && (str[3]=='(' || str[3]=='{' || str[3]==' ')) return TRUE;
    return FALSE;
}
int IsSubroutine(char* str){
    if(str[0]!='s' || str[1]!='u' || str[2]!='b') return 0;
    if(str[3]==' ') return 4;
    if(str[3]=='r' && str[4]=='o' && str[5]=='u'
       && str[6]=='t' && str[7]=='i' && str[8]=='n'
       && str[9]=='e' && str[10]==' ') return 11;
    return 0;
}

int FileSuffix(char* str)
{
    int i=0;
    while(str[i]!='\0') i++;
    while(i>0 && str[i]!='.') i--;
    if(i==0) return F_NOSUF;
    i++;
    if(strcmp(str+i, "gif")==0){
        return F_GIF;
    }else if(strcmp(str+i, "xpm")==0){
        return F_XPM;
    }else{
        return F_UNKNOWN;
    }
}

char EscapedCharacter(char c)
{
    switch(c){
      case 'a':
        return '\a';
      case 'b':
        return '\b';
      case 'f':
        return '\f';
      case 'n':
        return '\n';
      case 'r':
        return '\r';
      case 't':
        return '\t';
      case 'v':
        return '\v';
      default:
        return c;
    }
}

char* ExpandHomeDirectory(char* str)
{
    int i, l;
    char usr[64];
    char* dir;
    char* ret=NULL;
    struct passwd* pswd;
    if(str[0]!='~') return str;
    for(i=1; (str[i]!='\0' && str[i]!='/'); i++) usr[i-1]=str[i];
    usr[i-1]='\0';
    if(usr[0]=='\0'){
        dir=getenv("HOME");
    }else if((pswd=getpwnam(usr))){
        dir=pswd->pw_dir;
    }else{
        return str;
    }
    l=strlen(dir)+strlen(str+i)+2;
    ret=(char*)malloc(l*sizeof(char));
    sprintf(ret, "%s/%s", dir, str+i);
    return ret;
}




void LimitExpand(double* min, double* max){
    double size;
    size=*max-*min;
    if(size<=0.0) return;
    *min=*min-0.1*size;
    *max=*max+0.1*size;
}

List* Split(char c, char* str)
{
    char key[2];
    char* buf;
    List* list;
    Element* el;
    key[0]=c;  key[1]='\0';
    list=ToList(NULL);
    StripSpaces(&str);
    buf=str;
    while(*str!='\0'){
        if(*str==c){
            *str='\0';
            StripSpaces(&buf);
            el=ToElement(buf);
            if(el && el->type==VAR) el->type=STR;
            AddToList(list, el);
            str++;
            StripSpaces(&str);
            buf=str;
        }else{
            str++;
        }
    }
    if(*buf!='\0'){
        StripSpaces(&buf);
        el=ToElement(buf);
        if(el && el->type==VAR) el->type=STR;
        AddToList(list, el);
    }
    return list;
}

void SPrF(char* buf, char* form, Element* el)
{
    char type=0;
    int i;
    for(i=0; (!type && form[i]!='\0'); i++){
        switch(form[i]){
          case 'd':
          case 'i':
          case 'o':
          case 'u':
          case 'x':
          case 'X':
            type='d';
            break;
          case 'f':
          case 'e':
          case 'E':
          case 'g':
          case 'G':
            type='f';
            break;
          case 'c':
            type='c';
            break;
          case 's':
            type='s';
            break;
        }
    }
    if(type=='d'){
        sprintf(buf, form, ToInteger(el));
    }else if(type=='f'){
        sprintf(buf, form, ToDouble(el));
    }else if(type=='c'){
        int s;
        s=ToInteger(el);
        sprintf(buf, form, s);
    }else if(type=='s'){
        char* s;
        s=ToString(el);
        sprintf(buf, form, s);
        if(!el || el->type!=STR) free(s);
    }else{
        strcpy(buf, form);
    }
}
char* SPrintF(char* form, int n, Element* els[])
{
    int i=0, nbuf=256;
    char buf[256];
    char* f;
    char* str;
    char* ret;
    Element* el;
    f=form;
    while(*f!='%' && *f!='\0') f++;
    if(*f=='\0'){
        ret=(char*)malloc((strlen(form)+1)*sizeof(char));
        strcpy(ret, form);
        return ret;
    }
    *f='\0';
    if(nbuf<strlen(form)+1) nbuf+=256;
    str=(char*)malloc(nbuf*sizeof(char));
    strcpy(str, form);
    *f='%';
    form=f;
    while(*f!='\0'){
        f++;
        while(*f!='%' && *f!='\0') f++;
        if(i<n){
            el=els[i];  i++;
        }else{
            el=NULL;
        }
        if(*f=='\0'){
            SPrF(buf, form, el);
        }else{
            *f='\0';
            SPrF(buf, form, el);
            *f='%';
        }
        form=f;
        if(nbuf<strlen(str)+strlen(buf)+1){
            nbuf+=256;
            str=(char*)realloc(str, nbuf*sizeof(char));
        }
        strcat(str, buf);
    }
    ret=(char*)malloc((strlen(str)+1)*sizeof(char));
    strcpy(ret, str);
    free(str);
    return ret;
}


void EchoError(char* msg){
    fprintf(stderr, "%s\n", msg);
}
void CommandError(char* com, char* msg){
    if(com) fprintf(stderr, "%s: ", com);
    fprintf(stderr, "%s\n", msg);
}
void ParseError(int line, char* msg){
    e_parse=TRUE;
    if(filename[0]!='\0'){
        if(firsterr) fprintf(stderr, "in file \"%s\":\n", filename);
        fprintf(stderr, "line %d: ", line);
    }
    fprintf(stderr, "%s\n", msg);
    firsterr=FALSE;
}
void ExecError(char* msg){
    fprintf(stderr, "%s\n", msg);
}
int ParseErrorOccured(){
    return e_parse;
}
void ClearParseError(){
    e_parse=FALSE;
}
void SetFirstError(){
    firsterr=TRUE;
}
void EchoHelp()
{
    printf("Usage: pointer [options] [file]\n");
    printf("  options:\n");
    printf("    -h, --help      show this help\n");
    printf("    -v, --version   show version number\n");
    printf("    -ps             ps mode\n");
}
void SetFileName(char* name){
    if(name){
        strcpy(filename, name);
    }else{
        filename[0]='\0';
    }
    firsterr=TRUE;
}
void GetFileName(char* name){
    strcpy(name, filename);
}
void SetSubName(char* name){
    if(name){
        strcpy(subname, name);
    }else{
        subname[0]='\0';
    }
}
void GetSubName(char* name){
    strcpy(name, subname);
}

char* DefaultString(){
    Element* el;
    el=VarNameToElement(V_String);
    if(el && el->type==STR) return el->string.val;
    return NULL;
}
FILE* DefaultFile(){
    Element* el;
    el=VarNameToElement(V_File);
    if(el && el->type==FP) return el->file.fp;
    return NULL;
}
Element* DefaultFileElement(){
    Element* el;
    el=VarNameToElement(V_File);
    if(el && el->type==FP) return el;
    return NULL;
}
PTRWindow* ActiveWindow(){
    Element* el;
    el=VarNameToElement(V_Window);
    if(el && el->type==WINDOW) return el->window.wp;
    return NULL;
}
PTRData* DefaultData(){
    Element* el;
    el=VarNameToElement(V_Data);
    if(el && el->type==DATA){
        UpdateDataElement((Data*)el);
        return el->data.dp;
    }
    return NULL;
}
